from typing import Tuple


def find_intermediate_color_rgba(
    lowcolor_rgbastr: str, highcolor_rgbastr: str, intermed: float
) -> str:
    """
    Returns the color at a given distance between two colors
    This function takes two RGBA colors along with a value 0 < intermed < 1 and
    returns an RGBA color that is interpolated between the two input colors.
    The RGBA colors are assumed to be JavaScript color strings in the
    format 'rgba(r, g, b, a)' where r,g and b are intenisties between 0 and 255
    and a is opacity in the range 0.0 to 1.0
    """

    # convert to tuple color
    lowcolor = _unlabel_rgba(lowcolor_rgbastr)
    highcolor = _unlabel_rgba(highcolor_rgbastr)

    diff_0 = float(highcolor[0] - lowcolor[0])
    diff_1 = float(highcolor[1] - lowcolor[1])
    diff_2 = float(highcolor[2] - lowcolor[2])
    diff_3 = float(highcolor[3] - lowcolor[3])

    inter_med_tuple = (
        lowcolor[0] + intermed * diff_0,
        lowcolor[1] + intermed * diff_1,
        lowcolor[2] + intermed * diff_2,
        lowcolor[3] + intermed * diff_3,
    )

    # back to an rgba string, e.g. rgba(30, 20, 10)
    inter_med_rgba = _label_rgba(inter_med_tuple)
    return inter_med_rgba


def _label_rgba(colors: Tuple[float, float, float, float]) -> str:
    """
    Takes tuple (a, b, c, d) and returns an rgba color 'rgba(a, b, c, d)'
    """
    return f"rgba({colors[0]}, {colors[1]}, {colors[2]}, {colors[3]})"


def _unlabel_rgba(colors: str) -> Tuple[float, ...]:
    """
    Takes rgba color(s) 'rgba(a, b, c, d)' and returns tuple(s) (a, b, c, d)
    This function takes either an 'rgba(a, b, c, d)' color or a list of
    such colors and returns the color tuples in tuple(s) (a, b, c, d)
    """
    str_vals = ""
    for index, _col in enumerate(colors):
        try:
            float(colors[index])
            str_vals = str_vals + colors[index]
        except ValueError:
            if colors[index] == "," or colors[index] == ".":
                str_vals = str_vals + colors[index]

    str_vals = str_vals + ","
    numbers = []
    str_num = ""
    for char in str_vals:
        if char != ",":
            str_num = str_num + char
        else:
            numbers.append(float(str_num))
            str_num = ""
    return tuple(numbers)



from typing import Any, Callable, Dict, Optional

import webviz_core_components as wcc
from dash import ALL, Input, Output, State, callback, callback_context, no_update
from dash.exceptions import PreventUpdate

from webviz_subsurface._models import InplaceVolumesModel
from webviz_subsurface._utils.formatting import printable_int_list

from ..utils.utils import update_relevant_components


# pylint: disable=too-many-statements,too-many-arguments
def selections_controllers(
    get_uuid: Callable, volumemodel: InplaceVolumesModel
) -> None:
    @callback(
        Output(get_uuid("selections"), "data"),
        Input({"id": get_uuid("selections"), "tab": ALL, "selector": ALL}, "value"),
        Input(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": ALL},
            "value",
        ),
        Input(
            {"id": get_uuid("selections"), "tab": ALL, "settings": "Colorscale"},
            "colorscale",
        ),
        Input(get_uuid("initial-load-info"), "data"),
        State(get_uuid("page-selected"), "data"),
        State(get_uuid("tabs"), "value"),
        State(get_uuid("selections"), "data"),
        State({"id": get_uuid("selections"), "tab": ALL, "selector": ALL}, "id"),
        State(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": ALL}, "id"
        ),
    )
    def _update_selections(
        selectors: list,
        filters: list,
        colorscale: str,
        initial_load: dict,
        selected_page: str,
        selected_tab: str,
        previous_selection: dict,
        selector_ids: list,
        filter_ids: list,
    ) -> dict:
        ctx = callback_context.triggered[0]
        if ctx["prop_id"] == ".":
            raise PreventUpdate

        if previous_selection is None:
            previous_selection = {}

        page_selections = {
            id_value["selector"]: values
            for id_value, values in zip(selector_ids, selectors)
            if id_value["tab"] == selected_tab
        }
        page_selections["filters"] = {
            id_value["selector"]: values
            for id_value, values in zip(filter_ids, filters)
            if id_value["tab"] == selected_tab
        }

        page_selections.update(Colorscale=colorscale[0] if colorscale else None)
        page_selections.update(ctx_clicked=ctx["prop_id"])

        # check if a page needs to be updated due to page refresh or
        # change in selections/filters
        if initial_load[selected_page]:
            page_selections.update(update=True)
        else:
            equal_list = []
            for selector, values in page_selections.items():
                if selector != "ctx_clicked":
                    equal_list.append(
                        values == previous_selection[selected_page][selector]
                    )
            page_selections.update(update=not all(equal_list))

        previous_selection[selected_page] = page_selections
        return previous_selection

    @callback(
        Output(get_uuid("initial-load-info"), "data"),
        Input(get_uuid("page-selected"), "data"),
        Input({"id": get_uuid("selections"), "tab": ALL, "selector": ALL}, "value"),
        Input(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": ALL},
            "value",
        ),
        State(get_uuid("initial-load-info"), "data"),
    )
    def _store_initial_load_info(
        page_selected: str,
        _selectors_changed: list,
        _filters_changed: list,
        initial_load: dict,
    ) -> Dict[str, bool]:
        """
        Store info (True/False) reagarding if a page is initally loaded.
        Updating filters or selectors will set the value to False
        """
        if initial_load is None:
            initial_load = {}
        initial_load[page_selected] = page_selected not in initial_load
        return initial_load

    @callback(
        Output(
            {"id": get_uuid("selections"), "tab": "voldist", "selector": ALL},
            "disabled",
        ),
        Output(
            {"id": get_uuid("selections"), "tab": "voldist", "selector": ALL}, "value"
        ),
        Output(
            {"id": get_uuid("selections"), "tab": "voldist", "selector": ALL}, "options"
        ),
        Input(
            {"id": get_uuid("selections"), "tab": "voldist", "selector": "Plot type"},
            "value",
        ),
        Input(get_uuid("page-selected"), "data"),
        Input(
            {"id": get_uuid("selections"), "tab": "voldist", "selector": "Color by"},
            "value",
        ),
        State(
            {"id": get_uuid("selections"), "tab": "voldist", "selector": "bottom_viz"},
            "options",
        ),
        State(
            {"id": get_uuid("selections"), "tab": "voldist", "selector": ALL}, "value"
        ),
        State({"id": get_uuid("selections"), "tab": "voldist", "selector": ALL}, "id"),
        State(get_uuid("selections"), "data"),
        State(get_uuid("tabs"), "value"),
    )
    # pylint: disable=too-many-locals
    def _plot_options(
        plot_type: str,
        selected_page: str,
        selected_color_by: list,
        visualization_options: list,
        selector_values: list,
        selector_ids: list,
        previous_selection: Optional[dict],
        selected_tab: str,
    ) -> tuple:
        ctx = callback_context.triggered[0]
        if (
            selected_tab != "voldist"
            or ("Color by" in ctx["prop_id"] and plot_type not in ["box", "bar"])
            or previous_selection is None
        ):
            raise PreventUpdate

        selections: Any = (
            previous_selection.get(selected_page)
            if "page-selected" in ctx["prop_id"] and selected_page in previous_selection
            else {
                id_value["selector"]: values
                for id_value, values in zip(selector_ids, selector_values)
            }
        )

        selectors_disable_in_pages: Dict[str, list] = {
            "Plot type": ["per_zr", "conv"],
            "Y Response": ["per_zr", "conv"],
            "X Response": [],
            "Color by": ["conv"],
            "Subplots": ["per_zr"],
        }

        settings: Dict[str, dict] = {}
        for selector, disable_in_pages in selectors_disable_in_pages.items():
            disable = selected_page in disable_in_pages or (
                selector == "Y Response"
                and selections["Plot type"] in ["distribution", "histogram"]
            )
            value = None if disable else selections.get(selector)

            settings[selector] = {"disable": disable, "value": value}

        # update dropdown options based on plot type
        if settings["Plot type"]["value"] == "scatter":
            y_elm = x_elm = (
                volumemodel.responses + volumemodel.selectors + volumemodel.parameters
            )
        elif settings["Plot type"]["value"] in ["box", "bar"]:
            y_elm = x_elm = volumemodel.responses + volumemodel.selectors
            if selections.get("Y Response") is None:
                settings["Y Response"]["value"] = selected_color_by
        else:
            y_elm = volumemodel.selectors
            x_elm = volumemodel.responses

        colorby_elm = (
            list(volumemodel.dataframe.columns) + volumemodel.parameters
            if settings["Plot type"]["value"] == "scatter"
            else [x for x in volumemodel.selectors if x != "REAL"]
        )
        settings["Y Response"]["options"] = [
            {"label": elm, "value": elm} for elm in y_elm
        ]
        settings["X Response"]["options"] = [
            {"label": elm, "value": elm} for elm in x_elm
        ]
        if (
            settings["X Response"]["value"] is not None
            and settings["X Response"]["value"] not in x_elm
        ):
            settings["X Response"]["value"] = x_elm[0]

        settings["Color by"]["options"] = [
            {"label": elm, "value": elm} for elm in colorby_elm
        ]

        # disable vizualisation radioitem for some pages
        for x in visualization_options:
            x["disabled"] = selected_page != "custom"

        settings["bottom_viz"] = {
            "options": visualization_options,
            "value": "none"
            if selected_page != "custom"
            else selections.get("bottom_viz"),
        }

        return tuple(
            update_relevant_components(
                id_list=selector_ids,
                update_info=[
                    {
                        "new_value": values.get(prop, no_update),
                        "conditions": {"selector": selector},
                    }
                    for selector, values in settings.items()
                ],
            )
            for prop in ["disable", "value", "options"]
        )

    @callback(
        Output(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "undef"},
            "multi",
        ),
        Output(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "undef"},
            "value",
        ),
        Input(get_uuid("page-selected"), "data"),
        Input({"id": get_uuid("selections"), "tab": ALL, "selector": ALL}, "value"),
        State({"id": get_uuid("selections"), "tab": ALL, "selector": ALL}, "id"),
        State(get_uuid("tabs"), "value"),
        State(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "undef"},
            "options",
        ),
        State(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "undef"},
            "multi",
        ),
        State(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "undef"},
            "id",
        ),
    )
    # pylint: disable=too-many-locals
    def _update_filter_options(
        selected_page: str,
        selectors: list,
        selector_ids: list,
        selected_tab: str,
        filter_options: list,
        filter_multi: list,
        filter_ids: list,
    ) -> tuple:

        page_selections = {
            id_value["selector"]: values
            for id_value, values in zip(selector_ids, selectors)
            if id_value["tab"] == selected_tab
        }
        page_filter_settings = {
            id_value["selector"]: {"options": options, "multi": multi}
            for id_value, options, multi in zip(
                filter_ids, filter_options, filter_multi
            )
            if id_value["tab"] == selected_tab
        }

        selected_data = []
        if selected_tab == "voldist":
            selected_data = [
                page_selections[x]
                for x in ["Color by", "Subplots", "X Response", "Y Response"]
            ]
        if selected_tab == "table" and page_selections["Group by"] is not None:
            selected_data = page_selections["Group by"]
        if selected_tab == "tornado":
            selected_data = ["SENSNAME_CASE", page_selections["Subplots"]]
        if selected_tab == "ens-comp":
            selected_data = ["SENSNAME_CASE", "ENSEMBLE"]
        if selected_tab == "src-comp":
            selected_data = ["SOURCE"]

        # set "SENSNAME_CASE" multi also if "SENSNAME" OR "SENSCASE" is selected
        if any(senscol in selected_data for senscol in ("SENSNAME", "SENSCASE")):
            selected_data.append("SENSNAME_CASE")

        output = {}
        for selector in ["SOURCE", "ENSEMBLE", "SENSNAME_CASE"]:
            if selector not in page_filter_settings:
                continue
            options = [x["value"] for x in page_filter_settings[selector]["options"]]
            multi = selector in selected_data
            selector_is_multi = page_filter_settings[selector]["multi"]
            if not multi and selector_is_multi:
                values = [
                    "rms_seed"
                    if selector == "SENSNAME_CASE" and "rms_seed" in options
                    else options[0]
                ]
            elif multi and not selector_is_multi:
                values = options
            else:
                multi = values = no_update
            output[selector] = {"multi": multi, "values": values}

        # filter tornado on correct fluid based on volume response chosen
        output["FLUID_ZONE"] = {}
        if selected_page == "torn_bulk_inplace":
            output["FLUID_ZONE"] = {
                "values": ["oil" if page_selections["Response"] == "STOIIP" else "gas"]
            }

        return (
            update_relevant_components(
                id_list=filter_ids,
                update_info=[
                    {
                        "new_value": values.get("multi", no_update),
                        "conditions": {"tab": selected_tab, "selector": selector},
                    }
                    for selector, values in output.items()
                ],
            ),
            update_relevant_components(
                id_list=filter_ids,
                update_info=[
                    {
                        "new_value": values.get("values", no_update),
                        "conditions": {"tab": selected_tab, "selector": selector},
                    }
                    for selector, values in output.items()
                ],
            ),
        )

    @callback(
        Output(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "REAL"},
            "value",
        ),
        Output(
            {"id": get_uuid("filters"), "tab": ALL, "element": "real_text"}, "children"
        ),
        Input({"id": get_uuid("filters"), "tab": ALL, "component_type": ALL}, "value"),
        State(get_uuid("tabs"), "value"),
        State({"id": get_uuid("filters"), "tab": ALL, "component_type": ALL}, "id"),
        State({"id": get_uuid("filters"), "tab": ALL, "element": "real_text"}, "id"),
        State(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "REAL"},
            "id",
        ),
    )
    def _update_realization_filter_and_text(
        reals: list,
        selected_tab: str,
        reals_ids: list,
        real_string_ids: list,
        real_filter_id: list,
    ) -> tuple:
        """Callback that updates the selected relization info and text"""
        if selected_tab == "fipqc":
            raise PreventUpdate

        index = [x["tab"] for x in reals_ids].index(selected_tab)
        real_list = [int(real) for real in reals[index]]

        if reals_ids[index]["component_type"] == "range":
            real_list = list(range(real_list[0], real_list[1] + 1))

        return (
            update_relevant_components(
                id_list=real_filter_id,
                update_info=[
                    {"new_value": real_list, "conditions": {"tab": selected_tab}}
                ],
            ),
            update_relevant_components(
                id_list=real_string_ids,
                update_info=[
                    {
                        "new_value": printable_int_list(real_list),
                        "conditions": {"tab": selected_tab},
                    }
                ],
            ),
        )

    @callback(
        Output(
            {
                "id": get_uuid("filters"),
                "tab": ALL,
                "element": "real-slider-wrapper",
            },
            "children",
        ),
        Input(
            {
                "id": get_uuid("filters"),
                "tab": ALL,
                "element": "real-selector-option",
            },
            "value",
        ),
        State(get_uuid("selections"), "data"),
        State(get_uuid("page-selected"), "data"),
        State(get_uuid("tabs"), "value"),
        State(
            {
                "id": get_uuid("filters"),
                "tab": ALL,
                "element": "real-selector-option",
            },
            "id",
        ),
        State(
            {
                "id": get_uuid("filters"),
                "tab": ALL,
                "element": "real-slider-wrapper",
            },
            "id",
        ),
    )
    def _update_realization_selected_info(
        input_selectors: list,
        selections: dict,
        selected_page: str,
        selected_tab: str,
        input_ids: list,
        wrapper_ids: list,
    ) -> list:
        if selected_tab == "fipqc":
            raise PreventUpdate

        reals = volumemodel.realizations
        prev_selection = (
            selections[selected_page]["filters"].get("REAL", [])
            if selections is not None and selected_page in selections
            else None
        )
        selected_component = [
            value
            for id_value, value in zip(input_ids, input_selectors)
            if id_value["tab"] == selected_tab
        ][0]
        selected_reals = prev_selection if prev_selection is not None else reals

        component = (
            wcc.RangeSlider(
                id={
                    "id": get_uuid("filters"),
                    "tab": selected_tab,
                    "component_type": selected_component,
                },
                value=[min(selected_reals), max(selected_reals)],
                min=min(reals),
                max=max(reals),
                marks={str(i): {"label": str(i)} for i in [min(reals), max(reals)]},
            )
            if selected_component == "range"
            else wcc.SelectWithLabel(
                id={
                    "id": get_uuid("filters"),
                    "tab": selected_tab,
                    "component_type": selected_component,
                },
                options=[{"label": i, "value": i} for i in reals],
                value=selected_reals,
                size=min(20, len(reals)),
            )
        )
        return update_relevant_components(
            id_list=wrapper_ids,
            update_info=[{"new_value": component, "conditions": {"tab": selected_tab}}],
        )

    @callback(
        Output(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "region"},
            "value",
        ),
        Output(
            {"id": get_uuid("filters"), "wrapper": ALL, "tab": ALL, "type": "region"},
            "style",
        ),
        Input(
            {"id": get_uuid("filters"), "tab": ALL, "element": "region-selector"},
            "value",
        ),
        State(
            {"id": get_uuid("filters"), "tab": ALL, "selector": ALL, "type": "region"},
            "id",
        ),
        State(get_uuid("tabs"), "value"),
        State(get_uuid("page-selected"), "data"),
        State(get_uuid("selections"), "data"),
        State(
            {"id": get_uuid("filters"), "wrapper": ALL, "tab": ALL, "type": "region"},
            "id",
        ),
        State(
            {"id": get_uuid("filters"), "tab": ALL, "element": "region-selector"},
            "id",
        ),
        prevent_initial_call=True,
    )
    def update_region_filters(
        selected_reg_filter: list,
        reg_filter_ids: list,
        selected_tab: str,
        selected_page: str,
        selections: dict,
        wrapper_ids: list,
        reg_select_ids: list,
    ) -> tuple:
        """
        Callback to update the visible region filter between FIPNUM or ZONE/REGION.
        When changing, the active selection will be used to set the new selection.
        Note this callback will only be used for cases where each FIPNUM belongs to
        a unique ZONE and REGION.
        """
        selected = [
            value
            for id_value, value in zip(reg_select_ids, selected_reg_filter)
            if id_value["tab"] == selected_tab
        ]

        df = volumemodel.dataframe
        filters = selections[selected_page]["filters"]

        values = {}
        if selected[0] != "fipnum":
            values["FIPNUM"] = df["FIPNUM"].unique()
            for elm in ["REGION", "ZONE"]:
                values[elm] = df.loc[df["FIPNUM"].isin(filters["FIPNUM"])][elm].unique()

        else:
            values["REGION"] = df["REGION"].unique()
            values["ZONE"] = df["ZONE"].unique()
            mask = (df["REGION"].isin(filters["REGION"])) & (
                df["ZONE"].isin(filters["ZONE"])
            )
            values["FIPNUM"] = df.loc[mask]["FIPNUM"].unique()

        styles = {}
        styles["FIPNUM"] = {"display": "none" if selected[0] != "fipnum" else "block"}
        styles["REGION"] = {"display": "none" if selected[0] == "fipnum" else "block"}
        styles["ZONE"] = {"display": "none" if selected[0] == "fipnum" else "block"}

        return (
            update_relevant_components(
                id_list=reg_filter_ids,
                update_info=[
                    {
                        "new_value": value,
                        "conditions": {"selector": selector, "tab": selected_tab},
                    }
                    for selector, value in values.items()
                ],
            ),
            update_relevant_components(
                id_list=wrapper_ids,
                update_info=[
                    {
                        "new_value": style,
                        "conditions": {"wrapper": selector, "tab": selected_tab},
                    }
                    for selector, style in styles.items()
                ],
            ),
        )

    @callback(
        Output(
            {"id": get_uuid("selections"), "tab": "src-comp", "selector": "Ignore <"},
            "value",
        ),
        Input(
            {"id": get_uuid("selections"), "tab": "src-comp", "selector": "Response"},
            "value",
        ),
    )
    def _reset_ignore_value_source_comparison(_response_change: str) -> float:
        """reset ignore value when new response is selected"""
        return 0

    @callback(
        Output(
            {"id": get_uuid("selections"), "tab": "ens-comp", "selector": "Ignore <"},
            "value",
        ),
        Input(
            {"id": get_uuid("selections"), "tab": "ens-comp", "selector": "Response"},
            "value",
        ),
    )
    def _reset_ignore_value_ens_comparison(_response_change: str) -> float:
        """reset ignore value when new response is selected"""
        return 0



from pathlib import Path
from typing import Dict, List, Optional

import yaml


class ObservationModel:
    def __init__(
        self,
        observation_file: Path = None,
        observation_key: str = "general",
        remap_key: Dict[str, str] = None,
        remap_value: Dict[str, str] = None,
    ) -> None:
        self.observations = self.load_yaml(observation_file) if observation_file else {}
        self.observation_group = observation_key
        self.remap_key = remap_key if remap_key else {}
        self.remap_value = remap_value if remap_value else {}

    @staticmethod
    def load_yaml(yaml_file: Path) -> Dict:
        with open(Path(yaml_file), "r") as stream:
            return yaml.safe_load(stream)

    def get_attributes(self) -> List:
        return [
            attribute["key"]
            for attribute in self.observations.get(self.observation_group, [])
        ]

    def get_observations_for_attribute(
        self, attribute: str, value: str
    ) -> Optional[List]:
        for attr in self.observations.get(self.observation_group, []):
            if str(attr["key"]) == self.remap_key.get(attribute, attribute):
                observations = []
                for obs in attr.get("observations", []):
                    observations.append(
                        {
                            "value": obs.get("value", None),
                            "error": obs.get("error", None),
                            value: obs.get(self.remap_value.get(value, value), None),
                        }
                    )
                return observations
        return None



import glob
import pathlib
from typing import Dict, Optional

import jsonschema
import webviz_config
import yaml
from pkg_resources import DistributionNotFound, get_distribution

from webviz_subsurface._utils.user_defined_vector_definitions import (
    USER_DEFINED_VECTOR_DEFINITIONS_JSON_SCHEMA,
)
from webviz_subsurface._utils.vector_calculator import (
    PREDEFINED_EXPRESSIONS_JSON_SCHEMA,
    ConfigExpressionData,
)

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass


@webviz_config.SHARED_SETTINGS_SUBSCRIPTIONS.subscribe("scratch_ensembles")
def subscribe_scratch_ensembles(
    scratch_ensembles: dict, config_folder: pathlib.Path, portable: bool
) -> dict:
    if scratch_ensembles is not None:
        for ensemble_name, ensemble_path in scratch_ensembles.items():
            if not pathlib.Path(ensemble_path).is_absolute():
                scratch_ensembles[ensemble_name] = str(config_folder / ensemble_path)

            if not portable and not glob.glob(
                str(pathlib.Path(scratch_ensembles[ensemble_name]) / "OK")
            ):
                if not glob.glob(scratch_ensembles[ensemble_name]):
                    raise ValueError(
                        f"Ensemble {ensemble_name} is said to be located at {ensemble_path},"
                        " but that wildcard path does not give any matches."
                    )
                raise ValueError(
                    f"No realizations with a valid ERT 'OK' file found for ensemble "
                    f"{ensemble_name} located at {ensemble_path}. This can occur if an ERT job "
                    "has failed in all realizations, or none of the realizations have finished yet."
                )

    return scratch_ensembles


@webviz_config.SHARED_SETTINGS_SUBSCRIPTIONS.subscribe("predefined_expressions")
def subscribe_predefined_expressions(
    predefined_expressions: Optional[Dict[str, str]],
    config_folder: pathlib.Path,
    portable: bool,
) -> Dict[str, pathlib.Path]:

    output: Dict[str, pathlib.Path] = {}

    if predefined_expressions is None:
        return output

    for key, path in predefined_expressions.items():

        if not pathlib.Path(path).is_absolute():
            output[key] = config_folder / path

        if not portable:
            predefined_expressions_data: Dict[
                str, ConfigExpressionData
            ] = yaml.safe_load(output[key].read_text())

            try:
                jsonschema.validate(
                    instance=predefined_expressions_data,
                    schema=PREDEFINED_EXPRESSIONS_JSON_SCHEMA,
                )
            except jsonschema.exceptions.ValidationError as err:
                raise ValueError from err

    return output


@webviz_config.SHARED_SETTINGS_SUBSCRIPTIONS.subscribe(
    "user_defined_vector_definitions"
)
def subscribe_user_defined_vector_definitions(
    user_defined_vector_definitions: Optional[Dict[str, str]],
    config_folder: pathlib.Path,
    portable: bool,
) -> Dict[str, pathlib.Path]:

    output: Dict[str, pathlib.Path] = {}

    if user_defined_vector_definitions is None:
        return output

    for key, path in user_defined_vector_definitions.items():

        if not pathlib.Path(path).is_absolute():
            output[key] = config_folder / path

        if not portable:
            vector_definitions: Dict[str, str] = yaml.safe_load(output[key].read_text())

            try:
                jsonschema.validate(
                    instance=vector_definitions,
                    schema=USER_DEFINED_VECTOR_DEFINITIONS_JSON_SCHEMA,
                )
            except jsonschema.exceptions.ValidationError as err:
                raise ValueError from err

    return output



from typing import Callable, Dict, List, Union

from dash import dcc, html


def clientside_stores(
    get_uuid: Callable,
    initial_settings: Dict,
    realizations: List[Union[str, int]],
) -> html.Div:
    """Contains the clientside stores"""
    return html.Div(
        children=[
            dcc.Store(
                id=get_uuid("intersection-graph-data"), storage_type="session", data=[]
            ),
            dcc.Store(
                id=get_uuid("initial-intersection-graph-layout"),
                data=initial_settings.get("intersection_layout", {}),
            ),
            dcc.Store(
                id=get_uuid("realization-store"),
                data=initial_settings.get("intersection-data", {}).get(
                    "realizations", realizations
                ),
            ),
            dcc.Store(
                id={"id": get_uuid("map"), "element": "stored_polyline"},
                storage_type="session",
            ),
            dcc.Store(
                id={"id": get_uuid("map"), "element": "stored_xline"},
                storage_type="session",
            ),
            dcc.Store(
                id={"id": get_uuid("map"), "element": "stored_yline"},
                storage_type="session",
            ),
            dcc.Store(
                id=get_uuid("intersection-graph-layout"),
                storage_type="session",
            ),
            dcc.Store(
                id={
                    "id": get_uuid("intersection-data"),
                    "element": "stored_manual_update_options",
                },
                storage_type="session",
            ),
            dcc.Store(
                id=get_uuid("map-color-ranges"),
                storage_type="session",
            ),
        ]
    )



import logging
from pathlib import Path
from typing import Optional

import pandas as pd

# The fmu.ensemble dependency ecl is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from fmu.ensemble import ScratchEnsemble
except ImportError:
    pass

from webviz_subsurface._utils.perf_timer import PerfTimer

LOGGER = logging.getLogger(__name__)


def load_per_real_csv_file_using_fmu(
    ens_path: str, csv_file_rel_path: str
) -> pd.DataFrame:

    LOGGER.debug(f"load_per_real_csv_file_using_fmu() starting - {csv_file_rel_path}")
    timer = PerfTimer()

    scratch_ensemble = ScratchEnsemble("tempEnsName", ens_path, autodiscovery=True)
    df = scratch_ensemble.load_csv(csv_file_rel_path)

    LOGGER.debug(
        f"load_per_real_csv_file_using_fmu() finished in: {timer.elapsed_s():.2f}s"
    )

    return df


def load_ensemble_summary_csv_file(
    csv_file: Path, ensemble_filter: Optional[str]
) -> pd.DataFrame:

    LOGGER.debug(f"load_ensemble_summary_csv_file() starting - {csv_file}")
    timer = PerfTimer()

    df: pd.DataFrame = pd.read_csv(csv_file)

    if ensemble_filter is not None:
        if "ENSEMBLE" not in df.columns:
            raise ValueError(
                "Cannot filter on ensemble, no ENSEMBLE column exist in CSV file"
            )

        df = df[df["ENSEMBLE"] == ensemble_filter]

    if "ENSEMBLE" in df.columns:
        if df["ENSEMBLE"].nunique() > 1:
            raise KeyError("Input data contains more than one unique ensemble name")

        df = df.drop(columns="ENSEMBLE")

    LOGGER.debug(
        f"load_ensemble_summary_csv_file() finished in: {timer.elapsed_s():.2f}s"
    )

    return df



import abc
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Union

import xtgeo


class SurfaceStatistic(str, Enum):
    MEAN = "Mean"
    STDDEV = "StdDev"
    MINIMUM = "Minimum"
    MAXIMUM = "Maximum"
    P10 = "P10"
    P90 = "P90"


@dataclass(frozen=True)
class StatisticalSurfaceAddress:
    """Specifies a unique statistical surface in an ensemble"""

    attribute: str
    name: str
    datestr: Optional[str]
    statistic: SurfaceStatistic
    realizations: List[int]


@dataclass(frozen=True)
class SimulatedSurfaceAddress:
    """Specifies a unique simulated surface for a given ensemble realization"""

    attribute: str
    name: str
    datestr: Optional[str]
    realization: int


@dataclass(frozen=True)
class ObservedSurfaceAddress:
    """Represents a unique observed surface"""

    attribute: str
    name: str
    datestr: Optional[str]


# Type aliases used for signature readability
SurfaceAddress = Union[
    StatisticalSurfaceAddress, SimulatedSurfaceAddress, ObservedSurfaceAddress
]

# Class provides data for ensemble surfaces
class EnsembleSurfaceProvider(abc.ABC):
    @abc.abstractmethod
    def provider_id(self) -> str:
        """Returns string ID of the provider."""

    @abc.abstractmethod
    def attributes(self) -> List[str]:
        """Returns list of all available attributes."""

    @abc.abstractmethod
    def surface_names_for_attribute(self, surface_attribute: str) -> List[str]:
        """Returns list of all available surface names for a given attribute."""

    @abc.abstractmethod
    def surface_dates_for_attribute(
        self, surface_attribute: str
    ) -> Optional[List[str]]:
        """Returns list of all available surface dates for a given attribute."""

    @abc.abstractmethod
    def realizations(self) -> List[int]:
        """Returns list of all available realizations."""

    @abc.abstractmethod
    def get_surface(
        self,
        address: SurfaceAddress,
    ) -> Optional[xtgeo.RegularSurface]:
        """Returns a surface for a given surface address"""

    # @abc.abstractmethod
    # def get_surface_bounds(self, surface: EnsembleSurfaceContext) -> List[float]:
    #     """Returns the bounds for a surface [xmin,ymin, xmax,ymax]"""

    # @abc.abstractmethod
    # def get_surface_value_range(self, surface: EnsembleSurfaceContext) -> List[float]:
    #     """Returns the value range for a given surface context [zmin, zmax]"""

    # @abc.abstractmethod
    # def get_surface_as_rgba(self, surface: EnsembleSurfaceContext) -> io.BytesIO:
    #     """Returns surface as a greyscale png RGBA with encoded elevation values
    #     in a bytestream"""



import json
from pathlib import Path
from typing import Callable, Dict, List, Tuple, Union

import pandas as pd
import webviz_core_components as wcc
from dash import Dash, html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.webviz_assets import WEBVIZ_ASSETS

import webviz_subsurface
from webviz_subsurface._components import ColorPicker
from webviz_subsurface._datainput.fmu_input import find_surfaces, get_realizations
from webviz_subsurface._models import SurfaceSetModel, WellSetModel
from webviz_subsurface._utils.webvizstore_functions import find_files, get_path

from ._tour_steps import generate_tour_steps
from .controllers import (
    open_dialogs,
    update_intersection,
    update_intersection_source,
    update_maps,
    update_realizations,
    update_uncertainty_table,
)
from .views import (
    clientside_stores,
    dialog,
    intersection_and_map_layout,
    intersection_data_layout,
    map_data_layout,
    realization_layout,
    uncertainty_table_layout,
)


class StructuralUncertainty(WebvizPluginABC):
    """Dashboard to analyze structural uncertainty results from FMU runs.

A cross-section along a well or from a polyline drawn interactively on a map.
Map views to compare two surfaces from e.g. two iterations.

Both individual realization surfaces and statistical surfaces can be plotted.

Wells are required. If a zonelog is provided formation tops are extracted
and plotted as markers along the well trajectory.

Customization of colors and initialization of plugin with predefined selections
is possible. See the `Arguments` sections for details.

!> This plugin follows the FMU standards for storing and naming surface files.
Surface files must be stored at `share/results/maps` for each ensemble,
and be named as `surfacename--surfaceattribute.gri`

---

* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`surface_attributes`:** List of surface attributes from surface filenames(FMU standard). \
All surface_attributes must have the same surface names.
 * **`surface_name_filter`:** List of the surface names (FMU standard) in stratigraphic order
* **`wellfolder`:** A folder with wells on RMS Well format. \
(absolute or relative to config file).
* **`wellsuffix`:** File suffix for wells in `wellfolder`.
* **`zonelog`:** Name of zonelog in `wellfiles` (displayed along well trajectory).
* **`mdlog`:** Name of mdlog in `wellfiles` (displayed along well trajectory).
* **`well_tvdmin`:** Truncate well trajectory values above this depth.
* **`well_tvdmax`:** Truncate well trajectory values below this depth.
* **`well_downsample_interval`:** Sampling interval used for coarsening a well trajectory
* **`calculate_percentiles`:** Only relevant for portable. Calculating P10/90 is
time consuming and is by default disabled to allow fast generation of portable apps. \
Activate to precalculate these percentiles for all realizations. \
* **`initial_settings`:** Configuration for initializing the plugin with various \
    properties set. All properties are optional.
    ```yaml
        initial_settings:
            intersection_data: # Data to populate the intersection view
                surface_attribute: ds_extracted_horizons  #name of active attribute
                surface_names: #list of active surfacenames
                    - topupperreek
                    - baselowerreek
                ensembles: #list of active ensembles
                    - iter-0
                    - iter-1
                calculation: #list of active calculations
                    - Mean
                    - Min
                    - Max
                    - Realizations
                    - Uncertainty envelope
                well: OP_6 #Active well
                realizations: #List of active realizations
                    - 0
                resolution: 20 # Horizontal distance between points in the intersection
                             # (Usually in meters)
                extension: 500 # Horizontal extension of the intersection
                depth_truncations: # Truncations to use for yaxis range
                    min: 1500
                    max: 3000
            colors: # Colors to use for surfaces in the intersection view specified
                    # for each ensemble
                topupperreek:
                    iter-0: '#2C82C9' #hex color code with apostrophies and hash prefix
            intersection_layout: # The full plotly layout
                                 # (https://plotly.com/python/reference/layout/) is
                                 # exposed to allow for customization of e.g. plot title
                                 # and axis ranges. A small example:
                yaxis:
                    title: True vertical depth [m]
                xaxis:
                    title: Lateral distance [m]
    ```
---

**Example files**

* [One file for surfacefiles](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/\
maps/topupperreek--ds_extracted_horizons.gri).

* [Wellfiles](https://github.com/equinor/webviz-subsurface-testdata/tree/master/\
observed_data/wells).

The surfacefiles are on a `Irap binary` format and can be investigated outside `webviz` using \
e.g. [xtgeo](https://xtgeo.readthedocs.io/en/latest/).
"""

    # pylint: disable=too-many-arguments, too-many-instance-attributes, too-many-locals
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensembles: list,
        surface_attributes: list,
        surface_name_filter: List[str] = None,
        wellfolder: Path = None,
        wellsuffix: str = ".w",
        zonelog: str = None,
        mdlog: str = None,
        well_tvdmin: Union[int, float] = None,
        well_tvdmax: Union[int, float] = None,
        well_downsample_interval: int = None,
        calculate_percentiles: bool = False,
        initial_settings: Dict = None,
    ):

        super().__init__()
        self._initial_settings = initial_settings if initial_settings else {}

        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent
            / "_assets"
            / "css"
            / "structural_uncertainty.css"
        )
        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent
            / "_assets"
            / "js"
            / "clientside_functions.js"
        )
        self._calculate_percentiles = calculate_percentiles
        self._wellfolder = wellfolder
        self._wellsuffix = wellsuffix
        self._wellfiles: List = []
        if wellfolder is not None:
            self._wellfiles = json.load(find_files(wellfolder, wellsuffix))

        self._well_set_model = WellSetModel(
            self._wellfiles,
            zonelog=zonelog,
            mdlog=mdlog,
            tvdmin=well_tvdmin,
            tvdmax=well_tvdmax,
            downsample_interval=well_downsample_interval,
        )
        self._use_wells = bool(self._wellfiles)
        if (
            self._initial_settings.get("intersection_data", {}).get("well")
            and not self._use_wells
        ):
            raise KeyError(
                "Well is specified in initial settings but no well data is found!"
            )
        self._surf_attrs = surface_attributes
        self._ensemble_paths = {
            ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
            for ens in ensembles
        }

        # Create a table of surface files
        surface_table = find_surfaces(self._ensemble_paths)
        # Filter on provided surface attributes
        surface_table = surface_table[surface_table["attribute"].isin(self._surf_attrs)]
        # Filter on provided surface names
        self._surfacenames = (
            list(surface_table["name"].unique())
            if surface_name_filter is None
            else surface_name_filter
        )
        surface_table = surface_table[surface_table["name"].isin(surface_name_filter)]

        if surface_table.empty:
            raise ValueError("No surfaces found with the given attributes")

        self.ensembles = list(surface_table["ENSEMBLE"].unique())
        for _, attr_df in surface_table.groupby("attribute"):

            if set(attr_df["name"].unique()) != set(self._surfacenames):
                raise ValueError(
                    "Surface attributes has different surfaces. This is not supported!"
                )

        self._surface_ensemble_set_model = {
            ens: SurfaceSetModel(surf_ens_df)
            for ens, surf_ens_df in surface_table.groupby("ENSEMBLE")
        }
        self._realizations = sorted(list(surface_table["REAL"].unique()))

        self._zonelog = zonelog
        colors = [
            "#1f77b4",  # muted blue
            "#ff7f0e",  # safety orange
            "#2ca02c",  # cooked asparagus green
            "#d62728",  # brick red
            "#9467bd",  # muted purple
            "#8c564b",  # chestnut brown
            "#e377c2",  # raspberry yogurt pink
            "#7f7f7f",  # middle gray
            "#bcbd22",  # curry yellow-green
            "#17becf",  # blue-teal
        ]
        self._surfacecolors = [
            {
                "surfacename": surfacename,
                "ensemble": ens,
                "COLOR": self._initial_settings.get("colors", {})
                .get(surfacename, {})
                .get(ens, colors[idx % len(colors)]),
            }
            for idx, surfacename in enumerate(self._surfacenames)
            for ens in self.ensembles
        ]
        self._color_picker = ColorPicker(
            app=app,
            uuid=self.uuid("colorpicker"),
            dframe=pd.DataFrame(self._surfacecolors),
        )
        self.first_surface_geometry = self._surface_ensemble_set_model[
            self.ensembles[0]
        ].first_surface_geometry
        self.set_callbacks(app)

    @property
    def tour_steps(self) -> List[Dict]:
        return generate_tour_steps(get_uuid=self.uuid)

    @property
    def layout(self) -> wcc.FlexBox:
        return html.Div(
            id=self.uuid("layout"),
            children=[
                clientside_stores(
                    get_uuid=self.uuid,
                    realizations=self._realizations,
                    initial_settings=self._initial_settings,
                ),
                wcc.FlexBox(
                    children=[
                        wcc.FlexColumn(
                            wcc.Frame(
                                style={
                                    "height": "91vh",
                                },
                                children=[
                                    html.Div(
                                        children=[
                                            wcc.Selectors(
                                                label="Intersection controls",
                                                children=intersection_data_layout(
                                                    get_uuid=self.uuid,
                                                    surface_attributes=self._surf_attrs,
                                                    surface_names=self._surfacenames,
                                                    ensembles=self.ensembles,
                                                    use_wells=self._use_wells,
                                                    well_names=self._well_set_model.well_names
                                                    if self._well_set_model
                                                    else [],
                                                    surface_geometry=self.first_surface_geometry,
                                                    initial_settings=self._initial_settings.get(
                                                        "intersection_data", {}
                                                    ),
                                                ),
                                            ),
                                            html.Div(
                                                id=self.uuid(
                                                    "surface-settings-wrapper"
                                                ),
                                                children=wcc.Selectors(
                                                    label="Map controls",
                                                    children=[
                                                        map_data_layout(
                                                            uuid=self.uuid(
                                                                "map-settings"
                                                            ),
                                                            surface_attributes=self._surf_attrs,
                                                            surface_names=self._surfacenames,
                                                            ensembles=self.ensembles,
                                                            realizations=self._realizations,
                                                            use_wells=self._use_wells,
                                                        )
                                                    ],
                                                ),
                                            ),
                                            wcc.Selectors(
                                                label="Filters",
                                                children=[
                                                    dialog.open_dialog_layout(
                                                        uuid=self.uuid("dialog"),
                                                        dialog_id="realization-filter",
                                                        title="Realization filter",
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            )
                        ),
                        wcc.FlexColumn(
                            flex=6,
                            children=intersection_and_map_layout(get_uuid=self.uuid),
                        ),
                    ]
                ),
                dialog.dialog_layout(
                    uuid=self.uuid("dialog"),
                    dialog_id="color",
                    title="Color settings",
                    size="lg",
                    children=[
                        html.Div(
                            children=[self._color_picker.layout],
                        ),
                    ],
                ),
                dialog.dialog_layout(
                    uuid=self.uuid("dialog"),
                    dialog_id="realization-filter",
                    title="Filter realizations",
                    children=[
                        realization_layout(
                            uuid=self.uuid("intersection-data"),
                            realizations=self._realizations,
                            value=self._initial_settings.get(
                                "intersection_data", {}
                            ).get("realizations", self._realizations),
                        ),
                        dialog.clear_all_apply_dialog_buttons(
                            uuid=self.uuid("dialog"), dialog_id="realization-filter"
                        ),
                    ],
                ),
                dialog.dialog_layout(
                    uuid=self.uuid("dialog"),
                    dialog_id="uncertainty-table",
                    title="Uncertainty table",
                    children=[
                        uncertainty_table_layout(
                            uuid=self.uuid("uncertainty-table"),
                        )
                    ],
                ),
            ],
        )

    def set_callbacks(self, app: Dash) -> None:
        open_dialogs(app=app, get_uuid=self.uuid)
        update_realizations(app=app, get_uuid=self.uuid)
        update_intersection(
            app=app,
            get_uuid=self.uuid,
            surface_set_models=self._surface_ensemble_set_model,
            well_set_model=self._well_set_model,
            zonelog=self._zonelog,
            color_picker=self._color_picker,
        )
        update_maps(
            app=app,
            get_uuid=self.uuid,
            surface_set_models=self._surface_ensemble_set_model,
            well_set_model=self._well_set_model,
        )
        update_uncertainty_table(
            app=app,
            get_uuid=self.uuid,
            surface_set_models=self._surface_ensemble_set_model,
            well_set_model=self._well_set_model,
        )
        update_intersection_source(
            app=app, get_uuid=self.uuid, surface_geometry=self.first_surface_geometry
        )

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        store_functions: List[Tuple[Callable, list]] = []
        if self._use_wells is not None:
            store_functions.extend(
                [(get_path, [{"path": fn}]) for fn in self._wellfiles]
            )

        store_functions.append(
            (
                find_surfaces,
                [
                    {
                        "ensemble_paths": self._ensemble_paths,
                        "suffix": "*.gri",
                        "delimiter": "--",
                    }
                ],
            )
        )
        calculations: List[str] = ["Mean", "StdDev", "Min", "Max"]
        if self._calculate_percentiles:
            calculations.extend(["P10", "P90"])
        for ens in self.ensembles:

            for calculation in calculations:
                store_functions.append(
                    self._surface_ensemble_set_model[
                        ens
                    ].webviz_store_statistical_calculation(calculation=calculation)
                )
            store_functions.append(
                self._surface_ensemble_set_model[
                    ens
                ].webviz_store_realization_surfaces()
            )
        store_functions.append(
            (
                get_realizations,
                [
                    {
                        "ensemble_paths": self._ensemble_paths,
                        "ensemble_set_name": "EnsembleSet",
                    }
                ],
            )
        )
        if self._wellfolder is not None:
            store_functions.append(
                (find_files, [{"folder": self._wellfolder, "suffix": self._wellsuffix}])
            )
        return store_functions



from typing import Optional

from webviz_subsurface._providers import VectorMetadata

from .from_timeseries_cumulatives import is_per_interval_or_per_day_vector


def get_simulation_line_shape(
    line_shape_fallback: str,
    vector: str,
    vector_metadata: Optional[VectorMetadata] = None,
) -> str:
    """Get simulation time series line shape based on vector metadata"""
    if is_per_interval_or_per_day_vector(vector):
        # These custom calculated vectors are valid forwards in time.
        return "hv"

    if vector_metadata is None:
        return line_shape_fallback
    if vector_metadata.is_rate:
        # Eclipse rate vectors are valid backwards in time.
        return "vh"
    if vector_metadata.is_total:
        return "linear"
    return line_shape_fallback



import json
import logging
from pathlib import Path
from typing import Dict, List, Optional

import xtgeo

from webviz_subsurface._utils.perf_timer import PerfTimer

from .well_provider import WellPath, WellProvider

LOGGER = logging.getLogger(__name__)


INV_KEY_REL_PATH = "rel_path"
INV_KEY_MD_LOGNAME = "md_logname"


class ProviderImplFile(WellProvider):
    def __init__(
        self, provider_id: str, provider_dir: Path, inventory: Dict[str, dict]
    ) -> None:
        self._provider_id = provider_id
        self._provider_dir = provider_dir
        self._inventory = inventory

    @staticmethod
    def write_backing_store(
        storage_dir: Path,
        storage_key: str,
        well_file_names: List[str],
        md_logname: Optional[str],
    ) -> None:

        timer = PerfTimer()

        # All data for this provider will be stored inside a sub-directory
        # given by the storage key
        provider_dir = storage_dir / storage_key
        LOGGER.debug(f"Writing well backing store to: {provider_dir}")
        provider_dir.mkdir(parents=True, exist_ok=True)

        inventory_dict: Dict[str, dict] = {}

        LOGGER.debug(f"Writing {len(well_file_names)} wells into backing store...")

        timer.lap_s()
        for file_name in well_file_names:
            well = xtgeo.well_from_file(wfile=file_name, mdlogname=md_logname)

            if well.mdlogname is None:
                try:
                    well.geometrics()
                except ValueError:
                    LOGGER.debug(f"Ignoring {well.name} as MD cannot be calculated")
                    continue

            print("well.mdlogname=", well.mdlogname)

            well_name = well.name
            rel_path = f"{well_name}.rmswell"
            # rel_path = f"{well_name}.hdf"

            dst_file = provider_dir / rel_path
            print("dst_file=", dst_file)
            well.to_file(wfile=dst_file, fformat="rmswell")
            # well.to_hdf(wfile=dst_file)

            inventory_dict[well_name] = {
                INV_KEY_REL_PATH: rel_path,
                INV_KEY_MD_LOGNAME: well.mdlogname,
            }

        et_copy_s = timer.lap_s()

        json_fn = provider_dir / "inventory.json"
        with open(json_fn, "w") as file:
            json.dump(inventory_dict, file)

        LOGGER.debug(
            f"Wrote well backing store in: {timer.elapsed_s():.2f}s ("
            f"copy={et_copy_s:.2f}s)"
        )

    @staticmethod
    def from_backing_store(
        storage_dir: Path,
        storage_key: str,
    ) -> Optional["ProviderImplFile"]:

        provider_dir = storage_dir / storage_key
        json_fn = provider_dir / "inventory.json"

        try:
            with open(json_fn, "r") as file:
                inventory = json.load(file)
        except FileNotFoundError:
            return None

        return ProviderImplFile(storage_key, provider_dir, inventory)

    def provider_id(self) -> str:
        return self._provider_id

    def well_names(self) -> List[str]:
        return sorted(list(self._inventory.keys()))

    def get_well_path(self, well_name: str) -> WellPath:
        well = self.get_well_xtgeo_obj(well_name)
        df = well.dataframe
        md_logname = well.mdlogname

        x_arr = df["X_UTME"].to_numpy()
        y_arr = df["Y_UTMN"].to_numpy()
        z_arr = df["Z_TVDSS"].to_numpy()
        md_arr = df[md_logname].to_numpy()

        return WellPath(x_arr=x_arr, y_arr=y_arr, z_arr=z_arr, md_arr=md_arr)

    def get_well_xtgeo_obj(self, well_name: str) -> xtgeo.Well:
        well_entry = self._inventory.get(well_name)
        if not well_entry:
            raise ValueError(f"Requested well name {well_name} not found")

        rel_fn = well_entry[INV_KEY_REL_PATH]
        md_logname = well_entry[INV_KEY_MD_LOGNAME]

        full_file_name = self._provider_dir / rel_fn
        well = xtgeo.well_from_file(
            wfile=full_file_name, fformat="rmswell", mdlogname=md_logname
        )

        return well



from typing import Callable

import webviz_core_components as wcc
from dash import html

from ..models import ParametersModel
from .selector_view import ensemble_selector, filter_parameter, sortby_selector


def selector_view(get_uuid: Callable, parametermodel: ParametersModel) -> html.Div:
    return html.Div(
        [
            wcc.Selectors(
                label="Ensembles",
                children=[
                    ensemble_selector(
                        get_uuid=get_uuid,
                        ensembles=parametermodel.ensembles,
                        tab="qc",
                        id_string="ensemble-selector",
                        heading="Ensemble A:",
                        value=parametermodel.ensembles[0],
                    ),
                    ensemble_selector(
                        get_uuid=get_uuid,
                        ensembles=parametermodel.ensembles,
                        tab="qc",
                        id_string="delta-ensemble-selector",
                        heading="Ensemble B:",
                        value=parametermodel.ensembles[-1],
                    ),
                ],
            ),
            wcc.Selectors(
                label="Parameters",
                children=[
                    sortby_selector(get_uuid=get_uuid, value="Name"),
                    filter_parameter(
                        get_uuid=get_uuid,
                        parametermodel=parametermodel,
                        tab="qc",
                        value=[parametermodel.parameters[0]],
                    ),
                ],
            ),
        ],
    )


def parameter_qc_view(
    get_uuid: Callable, parametermodel: ParametersModel
) -> wcc.FlexBox:
    return wcc.FlexBox(
        children=[
            wcc.FlexColumn(
                wcc.Frame(
                    style={"height": "80vh"},
                    children=selector_view(
                        get_uuid=get_uuid, parametermodel=parametermodel
                    ),
                ),
            ),
            wcc.FlexColumn(
                flex=4,
                children=wcc.Frame(
                    color="white",
                    highlight=False,
                    style={"height": "80vh"},
                    children=[
                        wcc.RadioItems(
                            vertical=False,
                            id=get_uuid("property-qc-plot-type"),
                            options=[
                                {
                                    "label": "Distribution plots",
                                    "value": "distribution",
                                },
                                {"label": "Box plots", "value": "box"},
                                {"label": "Statistics table", "value": "table"},
                            ],
                            value="distribution",
                        ),
                        html.Div(
                            style={"height": "75vh", "margin-top": "20px"},
                            id=get_uuid("property-qc-wrapper"),
                        ),
                    ],
                ),
            ),
        ],
    )



from typing import Dict

import pytest

from webviz_subsurface._providers.ensemble_summary_provider.ensemble_summary_provider import (
    EnsembleSummaryProvider,
    VectorMetadata,
)
from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

from .mocks.ensemble_summary_provider_mock import EnsembleSummaryProviderMock

TEST_PROVIDER_DICT: Dict[str, EnsembleSummaryProvider] = {
    "First provider": EnsembleSummaryProviderMock.create_mock_with_first_dataset(),
    "Second provider": EnsembleSummaryProviderMock.create_mock_with_second_dataset(),
    "Third provider": EnsembleSummaryProviderMock.create_mock_with_third_dataset(),
}

TEST_INCONSISTENT_PROVIDER_DICT: Dict[str, EnsembleSummaryProvider] = {
    "First provider": EnsembleSummaryProviderMock.create_mock_with_first_dataset(),
    "Third provider": EnsembleSummaryProviderMock.create_mock_with_third_dataset(),
    "Inconsistent provider": EnsembleSummaryProviderMock.create_mock_with_inconsistent_dataset(),
}


def test_verify_consistent_vector_metadata() -> None:
    consistent_provider_set = EnsembleSummaryProviderSet(TEST_PROVIDER_DICT)
    inconsistent_provider_set = EnsembleSummaryProviderSet(
        TEST_INCONSISTENT_PROVIDER_DICT
    )

    # Expect no ValueError when verifying
    try:
        consistent_provider_set.verify_consistent_vector_metadata()
    except ValueError as err:
        pytest.fail(
            f"Expected successful validation for consistent_metadata_set. Exception: {err}"
        )

    # Expect ValueError when verifying
    with pytest.raises(ValueError):
        inconsistent_provider_set.verify_consistent_vector_metadata()
        pytest.fail(
            "Expected unsuccessful verification of consistent vector metadata for "
            "inconsistent_metadata_set"
        )


def test_create_union_of_vector_names_from_providers() -> None:
    # NOTE: Use explicit type annotation for EnsembleSummaryProviderMock as Dict is invariant.
    # See: https://mypy.readthedocs.io/en/latest/common_issues.html#variance
    provider_dict: Dict[str, EnsembleSummaryProvider] = {
        "First provider": EnsembleSummaryProviderMock.create_mock_with_first_dataset(),
        "Second provider": EnsembleSummaryProviderMock.create_mock_with_second_dataset(),
    }
    provider_set = EnsembleSummaryProviderSet(provider_dict)

    # Vector names from first and second mock - without duplicates sorted alphabetically
    expected_vector_names = [
        "FGIR",
        "WBHP:A1",
        "WBHP:A2",
        "WGOR:A1",
        "WGOR:A2",
        "WOPR:A2",
        "WOPT:A1",
        "WOPT:A2",
        "WWCT:A1",
        "WWCT:A2",
    ]

    # pylint: disable = protected-access
    created_vector_names = provider_set._create_union_of_vector_names_from_providers(
        list(provider_dict.values())
    )

    assert created_vector_names == expected_vector_names


def test_create_union_of_realizations_from_providers() -> None:
    # NOTE: Use explicit type annotation for EnsembleSummaryProviderMock as Dict is invariant.
    # See: https://mypy.readthedocs.io/en/latest/common_issues.html#variance
    provider_dict: Dict[str, EnsembleSummaryProvider] = {
        "First provider": EnsembleSummaryProviderMock.create_mock_with_first_dataset(),
        "Second provider": EnsembleSummaryProviderMock.create_mock_with_second_dataset(),
    }
    provider_set = EnsembleSummaryProviderSet(provider_dict)

    # Realizations from first and second mock
    expected_realizations = [1, 2, 3, 4, 5, 8]

    # pylint: disable = protected-access
    created_realizations = provider_set._create_union_of_realizations_from_providers(
        list(provider_dict.values())
    )

    assert created_realizations == expected_realizations


def test_items() -> None:
    provider_set = EnsembleSummaryProviderSet(TEST_PROVIDER_DICT)
    assert provider_set.items() == TEST_PROVIDER_DICT.items()


def test_names() -> None:
    provider_set = EnsembleSummaryProviderSet(TEST_PROVIDER_DICT)
    assert provider_set.provider_names() == list(TEST_PROVIDER_DICT.keys())


def test_provider() -> None:
    provider_set = EnsembleSummaryProviderSet(TEST_PROVIDER_DICT)

    first_provider = provider_set.provider("First provider")
    second_provider = provider_set.provider("Second provider")
    third_provider = provider_set.provider("Third provider")
    if not isinstance(first_provider, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected first provider "{first_provider}" to be type EnsembleSummaryProviderMock'
        )
    if not isinstance(second_provider, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected second provider "{second_provider}" to be type EnsembleSummaryProviderMock'
        )
    if not isinstance(third_provider, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected third provider "{third_provider}" to be type EnsembleSummaryProviderMock'
        )

    assert first_provider.get_dataset_name() == "First dataset"
    assert second_provider.get_dataset_name() == "Second dataset"
    assert third_provider.get_dataset_name() == "Third dataset"

    # Expect ValueError when getting invalid provider name
    with pytest.raises(ValueError):
        name = "Invalid Provider"
        provider_set.provider(name)
        pytest.fail(
            f'Expected getting provider "{name}" to fail, as it is not among provider names in set!'
        )


def test_all_providers() -> None:
    provider_set = EnsembleSummaryProviderSet(TEST_PROVIDER_DICT)
    all_providers = provider_set.all_providers()

    assert len(all_providers) == 3
    first_provider = all_providers[0]
    second_provider = all_providers[1]
    third_provider = all_providers[2]
    if not isinstance(first_provider, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected first provider "{first_provider}" to be type EnsembleSummaryProviderMock'
        )
    if not isinstance(second_provider, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected second provider "{second_provider}" to be type EnsembleSummaryProviderMock'
        )
    if not isinstance(third_provider, EnsembleSummaryProviderMock):
        pytest.fail(
            f'Expected third provider "{third_provider}" to be type EnsembleSummaryProviderMock'
        )

    assert first_provider.get_dataset_name() == "First dataset"
    assert second_provider.get_dataset_name() == "Second dataset"
    assert third_provider.get_dataset_name() == "Third dataset"


def test_all_realizations() -> None:
    provider_set = EnsembleSummaryProviderSet(TEST_PROVIDER_DICT)
    expected_realizations = [1, 2, 3, 4, 5, 7, 8]

    assert provider_set.all_realizations() == expected_realizations


def test_all_vector_names() -> None:
    provider_set = EnsembleSummaryProviderSet(TEST_PROVIDER_DICT)

    # Vector names from first, second and third mock - without duplicates sorted alphabetically
    expected_vector_names = [
        "FGIR",
        "FGIT",
        "WBHP:A1",
        "WBHP:A2",
        "WGOR:A1",
        "WGOR:A2",
        "WOPR:A1",
        "WOPR:A2",
        "WOPT:A1",
        "WOPT:A2",
        "WWCT:A1",
        "WWCT:A2",
    ]

    assert provider_set.all_vector_names() == expected_vector_names


def test_vector_metadata() -> None:
    provider_set = EnsembleSummaryProviderSet(TEST_PROVIDER_DICT)

    first_expected_metadata = VectorMetadata(
        unit="SM3/SM3",
        is_total=False,
        is_rate=True,
        is_historical=False,
        keyword="WGOR",
        wgname="A1",
        get_num=6,
    )

    second_expected_metadata = VectorMetadata(
        unit="",
        is_total=False,
        is_rate=True,
        is_historical=False,
        keyword="WWCT",
        wgname="A1",
        get_num=6,
    )

    third_expected_metadata = VectorMetadata(
        unit="SM3",
        is_total=True,
        is_rate=False,
        is_historical=False,
        keyword="FGIT",
        wgname=None,
        get_num=0,
    )

    assert provider_set.vector_metadata("WGOR:A1") == first_expected_metadata
    assert provider_set.vector_metadata("WWCT:A1") == second_expected_metadata
    assert provider_set.vector_metadata("FGIT") == third_expected_metadata
    assert provider_set.vector_metadata("Invalid Vector") is None


def test_vector_metadata_order() -> None:
    """The metadata returns first existing metadata, thereby inconsistent metadata
    will affect result based on the order of the providers in set.
    """

    # NOTE: Use explicit type annotation for EnsembleSummaryProviderMock as Dict is invariant.
    # See: https://mypy.readthedocs.io/en/latest/common_issues.html#variance
    first_provider_dict: Dict[str, EnsembleSummaryProvider] = {
        "First": EnsembleSummaryProviderMock.create_mock_with_first_dataset(),
        "Inconsistent": EnsembleSummaryProviderMock.create_mock_with_inconsistent_dataset(),
    }
    second_provider_dict: Dict[str, EnsembleSummaryProvider] = {
        "Inconsistent": EnsembleSummaryProviderMock.create_mock_with_inconsistent_dataset(),
        "First": EnsembleSummaryProviderMock.create_mock_with_first_dataset(),
    }
    first_provider_set = EnsembleSummaryProviderSet(first_provider_dict)
    second_provider_set = EnsembleSummaryProviderSet(second_provider_dict)

    # Metadata for first ensemble mock implementation
    first_ensemble_wwct_a1 = VectorMetadata(
        unit="",
        is_total=False,
        is_rate=True,
        is_historical=False,
        keyword="WWCT",
        wgname="A1",
        get_num=6,
    )
    first_ensemble_wgor_a2 = VectorMetadata(
        unit="SM3/SM3",
        is_total=False,
        is_rate=True,
        is_historical=False,
        keyword="WGOR",
        wgname="A2",
        get_num=7,
    )

    # Metadata for inconsistent ensemble mock implementation
    inconsistent_ensemble_wwct_a1 = VectorMetadata(
        unit="Invalid Unit",
        is_total=False,
        is_rate=False,
        is_historical=False,
        keyword="WWCT",
        wgname="A1",
        get_num=6,
    )
    inconsistent_ensemble_wgor_a2 = VectorMetadata(
        unit="SM3",
        is_total=False,
        is_rate=False,
        is_historical=True,
        keyword="WGOR",
        wgname="A2",
        get_num=7,
    )

    # First provider set should return metadata for the first ensemble mock implementation
    assert first_provider_set.vector_metadata("WWCT:A1") == first_ensemble_wwct_a1
    assert first_provider_set.vector_metadata("WGOR:A2") == first_ensemble_wgor_a2

    # Second provider set should return metadata for the inconsistent ensemble mock implementation
    assert (
        second_provider_set.vector_metadata("WWCT:A1") == inconsistent_ensemble_wwct_a1
    )
    assert (
        second_provider_set.vector_metadata("WGOR:A2") == inconsistent_ensemble_wgor_a2
    )



# pylint: disable=too-many-arguments
import logging
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

from webviz_config import WebvizPluginABC, WebvizSettings

from webviz_subsurface._models import WellAttributesModel
from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.ensemble_summary_provider_set_factory import (
    create_presampled_ensemble_summary_provider_set_from_paths,
)

from ._plugin_ids import PluginIds
from .shared_settings import Filter
from .views import (
    MisfitOptions,
    MisfitPerRealView,
    PlotSettingsCoverage,
    PlotSettingsHeatmap,
    PlotSettingsMisfit,
    ProdCoverageView,
    ProdHeatmapView,
)


class ProdMisfit(WebvizPluginABC):
    # pylint: disable=too-many-instance-attributes
    """Visualizes production data misfit at selected date(s).

    When not dealing with absolute value of differences, difference plots are
    represented as: (simulated - observed),
    i.e. negative values means sim is lower than obs and vice versa.

    **Features**
    * Visualization of prod misfit at selected time.
    * Visualization of prod coverage at selected time.
    * Heatmap representation of ensemble mean misfit for selected dates.

    ---
    * **`ensembles`:** Which ensembles in `shared_settings` to include.
    * **`rel_file_pattern`:** path to `.arrow` files with summary data.
    * **`sampling`:** Frequency for the data sampling.
    * **`well_attributes_file`:** Path to json file containing info of well attributes.
    The attribute category values can be used for filtering of well collections.
    * **`excl_name_startswith`:** Exclude wells that starts with this string.
    * **`excl_name_endswith`:** Exclude wells that ends with this string.
    * **`excl_name_contains`:** Exclude wells that contains this string.
    * **`phase_weights`:** Dict of "Oil", "Water" and "Gas" inverse weight factors that
    are included as weight option for misfit per real calculation.
    ---

    **Summary data**

    This plugin needs the following summary vectors to be stored with arrow format:
    * WOPT+WOPTH and/or WWPT+WWPTH and/or WGPT+WGPTH

    Summary files can be converted to arrow format with the `ECL2CSV` forward model.


    `well_attributes_file`: Optional json file with well attributes.
    The file needs to follow the format below. The categorical attributes \
    are completely flexible (user defined).
    ```json
    {
        "version" : "0.1",
        "wells" : [
        {
            "alias" : {
                "eclipse" : "A1"
            },
            "attributes" : {
                "structure" : "East",
                "welltype" : "producer"
            },
            "name" : "55_33-A-1"
        },
        {
            "alias" : {
                "eclipse" : "A5"
            },
            "attributes" : {
                "structure" : "North",
                "welltype" : "injector"
            },
            "name" : "55_33-A-5"
        },
        ]
    }
    ```
    """

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: list,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        sampling: str = Frequency.YEARLY.value,
        well_attributes_file: str = None,
        excl_name_startswith: list = None,
        excl_name_endswith: list = None,
        excl_name_contains: list = None,
        phase_weights: dict = None,
    ):
        # pylint: disable=too-many-statements

        super().__init__()

        if phase_weights is None:
            phase_weights = {"Oil": 1.0, "Water": 1.0, "Gas": 300.0}
        self.weight_reduction_factor_oil = phase_weights["Oil"]
        self.weight_reduction_factor_wat = phase_weights["Water"]
        self.weight_reduction_factor_gas = phase_weights["Gas"]

        if excl_name_startswith is None:
            excl_name_startswith = []
        excl_name_startswith = [str(element) for element in excl_name_startswith]
        if excl_name_endswith is None:
            excl_name_endswith = []
        excl_name_endswith = [str(element) for element in excl_name_endswith]
        if excl_name_contains is None:
            excl_name_contains = []
        excl_name_contains = [str(element) for element in excl_name_contains]

        # Must define valid frequency
        self._sampling = Frequency(sampling)

        ensemble_paths: Dict[str, Path] = {
            ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                ensemble_name
            ]
            for ensemble_name in ensembles
        }

        self._input_provider_set = (
            create_presampled_ensemble_summary_provider_set_from_paths(
                ensemble_paths, rel_file_pattern, self._sampling
            )
        )

        logging.debug("Created presampled provider_set.")

        self.ensemble_names = self._input_provider_set.provider_names()

        self.dates = {}
        self.realizations = {}
        self.wells = {}
        self.vectors = {}
        self.phases = {}

        self._well_attributes = (
            WellAttributesModel(
                self.ensemble_names[0],
                ensemble_paths[self.ensemble_names[0]],
                well_attributes_file,
            )
            if well_attributes_file is not None
            else None
        )

        for ens_name in self.ensemble_names:
            logging.debug(f"Working with: {ens_name}")
            ens_provider = self._input_provider_set.provider(ens_name)
            self.realizations[ens_name] = ens_provider.realizations()
            self.dates[ens_name] = ens_provider.dates(resampling_frequency=None)

            # from wopt/wwpt/wgpt: get lists of wells, vectors and phases
            # drop wells included in user input "excl_name" lists
            (
                self.wells[ens_name],
                self.vectors[ens_name],
                self.phases[ens_name],
            ) = _get_wells_vectors_phases(
                ens_provider.vector_names(),
                excl_name_startswith,
                excl_name_endswith,
                excl_name_contains,
            )

        # self.well_collections = _get_well_collections_from_attr(well_attrs, self.wells)
        self.well_collections = _get_well_collections_from_attr(
            self.wells, self._well_attributes
        )
        # ------------------------------------------------------------
        # calculations for settings
        self.all_dates, self.all_phases, self.all_wells, self.all_realizations = (
            [],
            [],
            [],
            [],
        )
        # pylint: disable=consider-iterating-dictionary
        for ens_name in self.ensemble_names:
            self.all_dates.extend(self.dates[ens_name])
            self.all_phases.extend(self.phases[ens_name])
            self.all_wells.extend(self.wells[ens_name])
            self.all_realizations.extend(self.realizations[ens_name])
        self.all_dates = list(sorted(set(self.all_dates)))
        self.all_phases = list(sorted(set(self.all_phases)))
        self.all_wells = list(sorted(set(self.all_wells)))
        self.all_realizations = list(sorted(set(self.all_realizations)))
        self.all_well_collection_names = []
        for collection_name in self.well_collections.keys():
            self.all_well_collection_names.append(collection_name)
        # --------------------------------------------------------------
        # add views, settings and stores

        self.add_store(
            PluginIds.Stores.SELECTED_ENSEMBLES, WebvizPluginABC.StorageType.SESSION
        )
        self.add_store(
            PluginIds.Stores.SELECTED_DATES, WebvizPluginABC.StorageType.SESSION
        )
        self.add_store(
            PluginIds.Stores.SELECTED_PHASE, WebvizPluginABC.StorageType.SESSION
        )
        self.add_store(
            PluginIds.Stores.SELECTED_WELLS, WebvizPluginABC.StorageType.SESSION
        )
        self.add_store(
            PluginIds.Stores.SELECTED_COMBINE_WELLS_COLLECTION,
            WebvizPluginABC.StorageType.SESSION,
        )
        self.add_store(
            PluginIds.Stores.SELECTED_WELL_COLLECTIONS,
            WebvizPluginABC.StorageType.SESSION,
        )
        self.add_store(
            PluginIds.Stores.SELECTED_REALIZATIONS, WebvizPluginABC.StorageType.SESSION
        )

        self.add_shared_settings_group(
            Filter(
                self.ensemble_names,
                self.all_dates,
                self.all_phases,
                self.all_wells,
                self.all_realizations,
                self.all_well_collection_names,
            ),
            PluginIds.SharedSettings.FILTER,
        )

        self.add_view(
            MisfitPerRealView(
                input_provider_set=self._input_provider_set,
                ens_vectors=self.vectors,
                ens_realizations=self.realizations,
                well_collections=self.well_collections,
                weight_reduction_factor_oil=self.weight_reduction_factor_oil,
                weight_reduction_factor_wat=self.weight_reduction_factor_wat,
                weight_reduction_factor_gas=self.weight_reduction_factor_gas,
            ),
            PluginIds.MisfitViews.PRODUCTION_MISFIT_PER_REAL,
        )
        self.add_view(
            ProdCoverageView(
                input_provider_set=self._input_provider_set,
                ens_vectors=self.vectors,
                ens_realizations=self.realizations,
                well_collections=self.well_collections,
            ),
            PluginIds.MisfitViews.WELL_PRODUCTION_COVERAGE,
        )
        self.add_view(
            ProdHeatmapView(
                input_provider_set=self._input_provider_set,
                ens_vectors=self.vectors,
                ens_realizations=self.realizations,
                well_collections=self.well_collections,
            ),
            PluginIds.MisfitViews.WELL_PRODUCTION_HEATMAP,
        )

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.view(PluginIds.MisfitViews.PRODUCTION_MISFIT_PER_REAL)
                .layout_element(MisfitPerRealView.Ids.MAIN_COLUMN)
                .get_unique_id(),
                "content": """Shows production misfit per realization.
                             Several ensembles can be shown at the same time.""",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.ENSEMBLE_SELECTOR),
                "content": """Select which ensembles to view graphs of.""",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.DATE_SELECTOR),
                "content": """Choose a single or several dates.""",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.PHASE_SELECTOR),
                "content": """Select what phases to be shown in the plot.""",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.WELL_SELECTOR),
                "content": """Select what wells to include in the data.""",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.COMBINE_WELL_AND_COLLECTION_AS),
                "content": """Combine the well and collection data as union or intersection.""",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.WELL_COLLECTION_SELECTOR),
                "content": """Choose what collection data to include.""",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.REALIZATION_SELECTOR),
                "content": """Choose how many realizations to include in the plot.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.PRODUCTION_MISFIT_PER_REAL)
                .settings_group(MisfitPerRealView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsMisfit.Ids.COLORBY),
                "content": """Color plot by phases, date or total misfit.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.PRODUCTION_MISFIT_PER_REAL)
                .settings_group(MisfitPerRealView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsMisfit.Ids.SORTING_RANKING),
                "content": """Rank data by ascending or descending value.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.PRODUCTION_MISFIT_PER_REAL)
                .settings_group(MisfitPerRealView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsMisfit.Ids.FIG_LAYOUT_HEIGHT),
                "content": """Select the size of the plot.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.PRODUCTION_MISFIT_PER_REAL)
                .settings_group(MisfitPerRealView.Ids.MISFIT_OPTIONS)
                .component_unique_id(MisfitOptions.Ids.MISFIT_WEIGHT),
                "content": """Select how to weigh misfits.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.PRODUCTION_MISFIT_PER_REAL)
                .settings_group(MisfitPerRealView.Ids.MISFIT_OPTIONS)
                .component_unique_id(MisfitOptions.Ids.MISFIT_EXPONENT),
                "content": """Choose between linear or square sum of misfit exponent.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.WELL_PRODUCTION_COVERAGE)
                .layout_element(ProdCoverageView.Ids.MAIN_COLUMN)
                .get_unique_id(),
                "content": """Shows well production coverage in a crossplot.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.WELL_PRODUCTION_COVERAGE)
                .settings_group(ProdCoverageView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsCoverage.Ids.COLORBY_GROUPING),
                "content": """Choose to have the ensembles in the plot overlay or side by side.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.WELL_PRODUCTION_COVERAGE)
                .settings_group(ProdCoverageView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsCoverage.Ids.SHOW_POINTS),
                "content": """Select how many points to show.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.WELL_PRODUCTION_HEATMAP)
                .layout_element(ProdHeatmapView.Ids.MAIN_COLUMN)
                .get_unique_id(),
                "content": """Shows cummulative misfit in heatmap.""",
            },
            {
                "id": self.view(PluginIds.MisfitViews.WELL_PRODUCTION_HEATMAP)
                .settings_group(ProdHeatmapView.Ids.PLOT_SETTINGS)
                .component_unique_id(PlotSettingsHeatmap.Ids.COLOR_RANGE_SCALING),
                "content": """Select the scale of the color range relative to max.""",
            },
        ]

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        return (
            [self._well_attributes.webviz_store]
            if self._well_attributes is not None
            else []
        )

    # ---------------------------------------------


# ------------------------------------------------------------------------
# support functions below here
# ------------------------------------------------------------------------

# ---------------------------
def _get_wells_vectors_phases(
    vector_names: List[str],
    excl_name_startswith: List[str],
    excl_name_endswith: List[str],
    excl_name_contains: List[str],
) -> Tuple[List, List, List]:
    """Return lists of wells, vectors and phases."""

    wells, vectors, drop_list = [], [], []
    phases = set()

    for vector in vector_names:
        if vector.startswith("WOPT:"):
            phases.add("Oil")
        elif vector.startswith("WWPT:"):
            phases.add("Water")
        elif vector.startswith("WGPT:"):
            phases.add("Gas")
        else:
            continue

        well = vector.split(":")[1]

        if _skip_well(
            well, excl_name_startswith, excl_name_endswith, excl_name_contains
        ):
            drop_list.append(well)
            continue

        if well not in wells:
            wells.append(well)
        if vector not in vectors:
            vectors.append(vector)

    wells, vectors = sorted(wells), sorted(vectors)

    if not vectors:
        RuntimeError("No WOPT, WWPT or WGPT vectors found.")

    if drop_list:
        logging.debug(
            "\nWells dropped based on config excl lists:\n"
            f"{list(sorted(set(drop_list)))}"
        )

    logging.debug(f"\nWells: {wells}")
    logging.debug(f"\nPhases: {phases}")
    logging.debug(f"\nVectors: {vectors}")

    return wells, vectors, list(phases)


def _skip_well(
    well: str,
    excl_name_startswith: List[str],
    excl_name_endswith: List[str],
    excl_name_contains: List[str],
) -> bool:
    """Check well name against exclude strings and return True if it should be skipped."""

    if well.startswith(tuple(excl_name_startswith)):
        return True
    if well.endswith(tuple(excl_name_endswith)):
        return True
    for excl in excl_name_contains:
        if excl in well:
            return True
    return False


# --------------------------------
def _get_well_collections_from_attr(
    wells: dict,
    well_attributes: Optional[WellAttributesModel],
) -> Dict[str, List[str]]:
    """Create well_collections dictionary. Then check
    well collections vs well lists. Any well not included in well collections is
    returned as Undefined."""

    all_wells = []
    for ens_wells in wells.values():
        all_wells.extend(ens_wells)
    all_wells = list(sorted(set(all_wells)))

    well_collections = {}

    if well_attributes is None:
        well_collections["Undefined"] = all_wells
        return well_collections

    # create well_collections dictionary from dataframe
    df_well_groups = well_attributes.dataframe_melted.dropna()
    df_cols = df_well_groups.columns
    if "WELL" not in df_cols or "VALUE" not in df_cols:
        RuntimeError(
            f"The {well_attributes.file_name} file must contain the columns"
            " 'WELL' and 'VALUE'"
        )
    for group in df_well_groups.groupby("VALUE"):
        well_collections[group[0]] = sorted(list(set(group[1].WELL.to_list())))

    undefined_wells = []
    all_collection_wells = []

    for collection_wells in well_collections.values():
        all_collection_wells.extend(collection_wells)
    all_collection_wells = list(set(all_collection_wells))
    undefined_wells = [well for well in all_wells if well not in all_collection_wells]
    if undefined_wells:
        well_collections["Undefined"] = undefined_wells
        logging.warning(
            "\nWells not included in any well collection ('Undefined'):"
            f"\n{undefined_wells}\n"
            f"Update the {well_attributes.file_name} file if they should be included"
        )

    return well_collections



from typing import List

import webviz_core_components as wcc
from dash import dcc, html


def map_data_layout(
    uuid: str,
    surface_attributes: List[str],
    surface_names: List[str],
    ensembles: List[str],
    realizations: List[int],
    use_wells: bool,
) -> html.Div:
    """Layout for the map data dialog"""
    return html.Div(
        children=[
            wcc.Selectors(
                label="Surface A",
                children=make_map_selectors(
                    uuid=uuid,
                    surface_attributes=surface_attributes,
                    surface_names=surface_names,
                    ensembles=ensembles,
                    realizations=realizations,
                    use_wells=use_wells,
                    map_id="map1",
                ),
            ),
            wcc.Selectors(
                label="Surface B",
                children=make_map_selectors(
                    uuid=uuid,
                    surface_attributes=surface_attributes,
                    surface_names=surface_names,
                    ensembles=ensembles,
                    realizations=realizations,
                    use_wells=use_wells,
                    map_id="map2",
                ),
            ),
            settings_layout(uuid=uuid),
        ]
    )


def make_map_selectors(
    uuid: str,
    map_id: str,
    surface_attributes: List[str],
    surface_names: List[str],
    ensembles: List[str],
    realizations: List[int],
    use_wells: bool,
) -> html.Div:
    return html.Div(
        children=[
            wcc.Dropdown(
                label="Surface attribute",
                id={"id": uuid, "map_id": map_id, "element": "surfaceattribute"},
                options=[{"label": val, "value": val} for val in surface_attributes],
                value=surface_attributes[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Surface name",
                id={"id": uuid, "map_id": map_id, "element": "surfacename"},
                options=[{"label": val, "value": val} for val in surface_names],
                value=surface_names[0],
                clearable=False,
            ),
            html.Div(
                style={"display": ("inline" if len(ensembles) > 1 else "none")},
                children=[
                    wcc.Dropdown(
                        label="Ensemble",
                        id={"id": uuid, "map_id": map_id, "element": "ensemble"},
                        options=[{"label": val, "value": val} for val in ensembles],
                        value=ensembles[0],
                        clearable=False,
                    ),
                ],
            ),
            wcc.Dropdown(
                label="Calculation/Realization",
                id={"id": uuid, "map_id": map_id, "element": "calculation"},
                options=[
                    {"label": val, "value": val}
                    for val in ["Mean", "StdDev", "Max", "Min", "P90", "P10"]
                    + [str(real) for real in realizations]
                ],
                value=str(realizations[0]),
                clearable=False,
            ),
            wcc.Checklist(
                id={"id": uuid, "map_id": map_id, "element": "options"},
                options=[
                    {"label": "Calculate well intersections", "value": "intersect_well"}
                ]
                if use_wells
                else [],
                value=[],
            ),
        ]
    )


def settings_layout(uuid: str) -> html.Details:
    return wcc.Selectors(
        open_details=False,
        label="⚙️ Settings",
        children=[
            wcc.Checklist(
                id={"id": uuid, "settings": "compute_diff"},
                options=[
                    {
                        "label": "Auto compute difference map",
                        "value": "compute_diffmap",
                    }
                ],
                value=["compute_diffmap"],
            ),
            html.Div(
                children=[
                    wcc.Label(
                        "Color ranges:",
                        style={"font-weight": "bold"},
                    ),
                    color_range_layout(uuid=uuid, map_id="map1"),
                    color_range_layout(uuid=uuid, map_id="map2"),
                    wcc.Checklist(
                        id={"id": uuid, "colors": "sync_range"},
                        options=[
                            {
                                "label": "Sync range on maps",
                                "value": "sync_range",
                            }
                        ],
                        value=[],
                    ),
                ],
            ),
        ],
    )


def color_range_layout(uuid: str, map_id: str) -> wcc.FlexBox:
    return wcc.FlexBox(
        style={"display": "flex", "align-items": "center"},
        children=[
            wcc.Label(
                "Surface A" if map_id == "map1" else "Surface B",
                style={"flex": 1, "minWidth": "40px"},
            ),
            dcc.Input(
                id={
                    "id": uuid,
                    "colors": f"{map_id}_clip_min",
                },
                style={"flex": 1, "minWidth": "40px"},
                type="number",
                value=None,
                debounce=True,
                placeholder="Min",
                persistence=True,
                persistence_type="session",
            ),
            dcc.Input(
                id={
                    "id": uuid,
                    "colors": f"{map_id}_clip_max",
                },
                style={"flex": 1, "minWidth": "40px"},
                type="number",
                value=None,
                debounce=True,
                placeholder="Max",
                persistence=True,
                persistence_type="session",
            ),
        ],
    )



from typing import Any, Dict, List

import plotly.graph_objects as go
import webviz_core_components as wcc
from dash import html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewABC, ViewElementABC
from webviz_subsurface_components import DeckGLMap


class MainView(ViewABC):
    class Ids(StrEnum):
        MAIN_ELEMENT = "main-element"

    def __init__(self, color_scales: List[Dict[str, Any]]):
        super().__init__("Main View")
        self._view_element = MapViewElement(color_scales)
        self.add_view_element(self._view_element, self.Ids.MAIN_ELEMENT)


class MapViewElement(ViewElementABC):
    class Ids(StrEnum):
        DECKGL_MAP = "deck-gl-map"
        DATE_SLIDER = "date-slider"
        DATE_WRAPPER = "date-wrapper"
        BAR_PLOT = "bar-plot"
        TIME_PLOT = "time-plot"

    def __init__(self, color_scales: List[Dict[str, Any]]) -> None:
        super().__init__()
        self._color_scales = color_scales

    def inner_layout(self) -> Component:
        return html.Div(
            [
                wcc.Frame(
                    # id=self.register_component_unique_id(LayoutElements.MAP_VIEW),
                    color="white",
                    highlight=False,
                    children=[
                        html.Div(
                            [
                                DeckGLMap(
                                    id=self.register_component_unique_id(
                                        self.Ids.DECKGL_MAP
                                    ),
                                    layers=[],
                                    coords={"visible": True},
                                    scale={"visible": True},
                                    toolbar={"visible": True},
                                    coordinateUnit="m",
                                    colorTables=self._color_scales,
                                ),
                            ],
                            style={
                                "padding": "1vh",
                                "height": "37vh",
                                "position": "relative",
                            },
                        ),
                        html.Div(
                            wcc.Slider(
                                id=self.register_component_unique_id(
                                    self.Ids.DATE_SLIDER
                                ),
                                step=None,
                                marks={0: ""},
                                value=0,
                            ),
                            id=self.register_component_unique_id(self.Ids.DATE_WRAPPER),
                        ),
                    ],
                    style={
                        "height": "47vh",
                    },
                ),
                wcc.Frame(
                    # id=get_uuid(LayoutElements.PLOT_VIEW),
                    style={
                        "height": "33vh",
                        "display": "flex",
                        "flexDirection": "row",
                        "justifyContent": "space-evenly",
                    },
                    children=SummaryGraphLayout(
                        self.register_component_unique_id(self.Ids.BAR_PLOT),
                        self.register_component_unique_id(self.Ids.TIME_PLOT),
                    ).children,
                ),
            ],
            style={
                "flex": 3,
                "display": "flex",
                "flexDirection": "column",
            },
        )


class SummaryGraphLayout(html.Div):
    def __init__(self, bar_plot_id: str, time_plot_id: str, **kwargs: Dict) -> None:
        super().__init__(
            children=[
                wcc.Graph(
                    id=bar_plot_id,
                    figure=go.Figure(),
                    config={
                        "displayModeBar": False,
                    },
                ),
                wcc.Graph(
                    id=time_plot_id,
                    figure=go.Figure(),
                    config={
                        "displayModeBar": False,
                    },
                ),
            ],
            **kwargs,
        )



import dataclasses
import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import numpy as np
from vtkmodules.util.numpy_support import vtk_to_numpy

# pylint: disable=no-name-in-module,
from vtkmodules.vtkCommonCore import reference, vtkIdList, vtkPoints

# pylint: disable=no-name-in-module,
from vtkmodules.vtkCommonDataModel import (
    vtkCellArray,
    vtkCellLocator,
    vtkExplicitStructuredGrid,
    vtkLine,
    vtkPlane,
    vtkPolyData,
    vtkUnstructuredGrid,
)

# Requires VTK 9.2
from vtkmodules.vtkFiltersCore import (
    vtkAppendPolyData,
    vtkClipPolyData,
    vtkExplicitStructuredGridCrop,
    vtkExplicitStructuredGridToUnstructuredGrid,
    vtkExtractCellsAlongPolyLine,
    vtkPlaneCutter,
    vtkUnstructuredGridToExplicitStructuredGrid,
)
from vtkmodules.vtkFiltersGeometry import vtkExplicitStructuredGridSurfaceFilter

from webviz_subsurface._utils.perf_timer import PerfTimer

# Requires updated xtgeo
from ._xtgeo_to_vtk_explicit_structured_grid import (
    xtgeo_grid_to_vtk_explicit_structured_grid,
)
from .ensemble_grid_provider import EnsembleGridProvider

LOGGER = logging.getLogger(__name__)

_GRID_VIZ_SERVICE_INSTANCE: Optional["GridVizService"] = None


@dataclass
class PropertySpec:
    prop_name: str
    prop_date: Optional[str]


@dataclass
class CellFilter:
    i_min: int
    i_max: int
    j_min: int
    j_max: int
    k_min: int
    k_max: int


@dataclass
class SurfacePolys:
    point_arr: np.ndarray
    poly_arr: np.ndarray


@dataclass
class PropertyScalars:
    value_arr: np.ndarray
    # min_value: float
    # max_value: float


@dataclass
class Ray:
    origin: List[float]
    end: List[float]
    # direction: List[float]


@dataclass
class PickResult:
    cell_index: int
    cell_i: int
    cell_j: int
    cell_k: int
    intersection_point: List[float]
    cell_property_value: Optional[float]


# =============================================================================
class GridWorker:
    # -----------------------------------------------------------------------------
    def __init__(self, full_esgrid: vtkExplicitStructuredGrid) -> None:
        self._full_esgrid = full_esgrid

        self._cached_cell_filter: Optional[CellFilter] = None
        self._cached_original_cell_indices: Optional[np.ndarray] = None

    # -----------------------------------------------------------------------------
    def get_full_esgrid(self) -> vtkExplicitStructuredGrid:
        return self._full_esgrid

    # -----------------------------------------------------------------------------
    def get_cached_original_cell_indices(
        self, cell_filter: Optional[CellFilter]
    ) -> Optional[np.ndarray]:
        if self._cached_original_cell_indices is None:
            return None

        if cell_filter == self._cached_cell_filter:
            return self._cached_original_cell_indices

        return None

    # -----------------------------------------------------------------------------
    def set_cached_original_cell_indices(
        self, cell_filter: Optional[CellFilter], original_cell_indices: np.ndarray
    ) -> None:
        # Make copy of the cell filter
        self._cached_cell_filter = dataclasses.replace(cell_filter)
        self._cached_original_cell_indices = original_cell_indices


# =============================================================================
class GridVizService:
    # -----------------------------------------------------------------------------
    def __init__(self) -> None:
        self._id_to_provider_dict: Dict[str, EnsembleGridProvider] = {}
        self._key_to_worker_dict: Dict[str, GridWorker] = {}

    # -----------------------------------------------------------------------------
    @staticmethod
    def instance() -> "GridVizService":
        # pylint: disable=global-statement,
        global _GRID_VIZ_SERVICE_INSTANCE
        if not _GRID_VIZ_SERVICE_INSTANCE:
            LOGGER.debug("Initializing GridVizService instance")
            _GRID_VIZ_SERVICE_INSTANCE = GridVizService()

        return _GRID_VIZ_SERVICE_INSTANCE

    # -----------------------------------------------------------------------------
    def register_provider(self, provider: EnsembleGridProvider) -> None:
        provider_id = provider.provider_id()
        LOGGER.debug(f"Adding grid provider with id={provider_id}")

        existing_provider = self._id_to_provider_dict.get(provider_id)
        if existing_provider:
            # Issue a warning if there already is a provider registered with the same
            # id AND if the actual provider instance is different.
            # This wil happen until the provider factory gets caching.
            if existing_provider is not provider:
                LOGGER.warning(
                    f"Provider with id={provider_id} ignored, the id is already present"
                )
                return

        self._id_to_provider_dict[provider_id] = provider

    # -----------------------------------------------------------------------------
    # pylint: disable=too-many-locals,
    def get_surface(
        self,
        provider_id: str,
        realization: int,
        property_spec: Optional[PropertySpec],
        cell_filter: Optional[CellFilter],
    ) -> Tuple[SurfacePolys, Optional[PropertyScalars]]:

        LOGGER.debug(
            f"Getting grid surface... "
            f"(provider_id={provider_id}, real={realization}, "
            f"{_property_spec_dbg_str(property_spec)}, "
            f"{_cell_filter_dbg_str(cell_filter)})"
        )
        timer = PerfTimer()

        provider = self._id_to_provider_dict.get(provider_id)
        if not provider:
            raise ValueError("Could not find provider")

        worker = self._get_or_create_grid_worker(provider_id, realization)
        if not worker:
            raise ValueError("Could not get grid worker")
        et_get_grid_worker_ms = timer.lap_ms()

        grid = worker.get_full_esgrid()

        if cell_filter:
            grid = _calc_cropped_grid(grid, cell_filter)
        et_crop_grid_ms = timer.lap_ms()

        polydata = _calc_grid_surface(grid)

        # !!!!!!
        # Need to watch out here, think these may go out of scope!
        points_np = vtk_to_numpy(polydata.GetPoints().GetData()).ravel()
        polys_np = vtk_to_numpy(polydata.GetPolys().GetData())
        original_cell_indices_np = vtk_to_numpy(
            polydata.GetCellData().GetAbstractArray("vtkOriginalCellIds")
        )

        surface_polys = SurfacePolys(point_arr=points_np, poly_arr=polys_np)
        et_calc_surf_ms = timer.lap_ms()

        property_scalars: Optional[PropertyScalars] = None
        if property_spec:
            raw_cell_vals = _load_property_values(provider, realization, property_spec)
            if raw_cell_vals is not None:
                mapped_cell_vals = raw_cell_vals[original_cell_indices_np]
                property_scalars = PropertyScalars(value_arr=mapped_cell_vals)
        et_read_and_map_scalars_ms = timer.lap_ms()

        worker.set_cached_original_cell_indices(cell_filter, original_cell_indices_np)

        LOGGER.debug(
            f"Got grid surface in {timer.elapsed_s():.2f}s "
            f"(get_grid_worker={et_get_grid_worker_ms}ms, "
            f"crop_grid={et_crop_grid_ms}ms, "
            f"calc_surf={et_calc_surf_ms}ms, "
            f"read_and_map_scalars={et_read_and_map_scalars_ms}ms, "
            f"provider_id={provider_id}, real={realization}, "
            f"{_property_spec_dbg_str(property_spec)}, "
            f"{_cell_filter_dbg_str(cell_filter)})"
        )

        return surface_polys, property_scalars

    # -----------------------------------------------------------------------------
    def get_mapped_property_values(
        self,
        provider_id: str,
        realization: int,
        property_spec: PropertySpec,
        cell_filter: Optional[CellFilter],
    ) -> Optional[PropertyScalars]:

        LOGGER.debug(
            f"Getting property values... "
            f"(provider_id={provider_id}, real={realization}, "
            f"{_property_spec_dbg_str(property_spec)}, "
            f"{_cell_filter_dbg_str(cell_filter)})"
        )
        timer = PerfTimer()

        provider = self._id_to_provider_dict.get(provider_id)
        if not provider:
            raise ValueError("Could not find provider")

        worker = self._get_or_create_grid_worker(provider_id, realization)
        if not worker:
            raise ValueError("Could not get grid worker")
        et_get_grid_worker_ms = timer.lap_ms()

        original_cell_indices_np = worker.get_cached_original_cell_indices(cell_filter)
        if original_cell_indices_np is None:
            # Must first generate the grid to get the original cell indices
            grid = worker.get_full_esgrid()
            if cell_filter:
                grid = _calc_cropped_grid(grid, cell_filter)

            polydata = _calc_grid_surface(grid)
            original_cell_indices_np = vtk_to_numpy(
                polydata.GetCellData().GetAbstractArray("vtkOriginalCellIds")
            )
            worker.set_cached_original_cell_indices(
                cell_filter, original_cell_indices_np
            )
        et_get_mapping_indices_ms = timer.lap_ms()

        raw_cell_vals = _load_property_values(provider, realization, property_spec)
        if raw_cell_vals is None:
            LOGGER.warning(
                f"No cell values found for "
                f"prop=({property_spec.prop_name}, {property_spec.prop_name})"
            )
            return None

        mapped_cell_vals = raw_cell_vals[original_cell_indices_np]
        et_read_and_map_scalars_ms = timer.lap_ms()

        LOGGER.debug(
            f"Got property values in {timer.elapsed_s():.2f}s "
            f"(get_grid_worker={et_get_grid_worker_ms}ms, "
            f"get_mapping_indices={et_get_mapping_indices_ms}ms, "
            f"read_and_map_scalars={et_read_and_map_scalars_ms}ms, "
            f"provider_id={provider_id}, real={realization}, "
            f"{_property_spec_dbg_str(property_spec)}, "
            f"{_cell_filter_dbg_str(cell_filter)})"
        )

        return PropertyScalars(value_arr=mapped_cell_vals)

    # -----------------------------------------------------------------------------
    # pylint: disable=too-many-locals,too-many-statements
    def cut_along_polyline(
        self,
        provider_id: str,
        realization: int,
        polyline_xy: List[float],
        property_spec: Optional[PropertySpec],
    ) -> Tuple[SurfacePolys, Optional[PropertyScalars]]:

        LOGGER.debug(
            f"Cutting along polyline... "
            f"(provider_id={provider_id}, real={realization})"
        )
        timer = PerfTimer()

        provider = self._id_to_provider_dict.get(provider_id)
        if not provider:
            raise ValueError("Could not find provider")

        worker = self._get_or_create_grid_worker(provider_id, realization)
        if not worker:
            raise ValueError("Could not get grid worker")

        esgrid = worker.get_full_esgrid()

        num_points_in_polyline = int(len(polyline_xy) / 2)

        ugrid = _vtk_esg_to_ug(esgrid)

        # !!!!!!!!!!!!!!
        # Requires VTK 9.2-ish
        # ugrid = _extract_intersected_ugrid(ugrid, polyline_xy, 10.0)

        cutter_alg = vtkPlaneCutter()
        cutter_alg.SetInputDataObject(ugrid)

        # cell_locator = vtkStaticCellLocator()
        # cell_locator.SetDataSet(esgrid)
        # cell_locator.BuildLocator()

        # box_clip_alg = vtkBoxClipDataSet()
        # box_clip_alg.SetInputDataObject(ugrid)

        append_alg = vtkAppendPolyData()
        et_setup_s = timer.lap_s()

        et_cut_s = 0.0
        et_clip_s = 0.0

        for i in range(0, num_points_in_polyline - 1):
            x_0 = polyline_xy[2 * i]
            y_0 = polyline_xy[2 * i + 1]
            x_1 = polyline_xy[2 * (i + 1)]
            y_1 = polyline_xy[2 * (i + 1) + 1]
            fwd_vec = np.array([x_1 - x_0, y_1 - y_0, 0.0])
            fwd_vec /= np.linalg.norm(fwd_vec)
            right_vec = np.array([fwd_vec[1], -fwd_vec[0], 0])

            # box_clip_alg.SetBoxClip(x_0, x_1, y_0, y_1, min_z, max_z)
            # box_clip_alg.Update()
            # clipped_ugrid = box_clip_alg.GetOutputDataObject(0)

            # polyline_bounds = _calc_polyline_bounds([x_0, y_0, x_1, y_1])
            # polyline_bounds.extend([min_z, max_z])
            # cell_ids = vtkIdList()
            # cell_locator.FindCellsWithinBounds(polyline_bounds, cell_ids)
            # print(f"{cell_ids.GetNumberOfIds()}  {polyline_bounds=}")

            plane = vtkPlane()
            plane.SetOrigin([x_0, y_0, 0])
            plane.SetNormal(right_vec)

            plane_0 = vtkPlane()
            plane_0.SetOrigin([x_0, y_0, 0])
            plane_0.SetNormal(fwd_vec)

            plane_1 = vtkPlane()
            plane_1.SetOrigin([x_1, y_1, 0])
            plane_1.SetNormal(-fwd_vec)

            cutter_alg.SetPlane(plane)
            cutter_alg.Update()

            cut_surface_polydata = cutter_alg.GetOutput()
            # print(f"{type(cut_surface_polydata)=}")
            et_cut_s += timer.lap_s()

            # Used vtkPolyDataPlaneClipper earlier, but it seems that it doesn't
            # maintain the original cell IDs that we need for the result mapping.
            # May want to check up on any performance degradation!
            clipper_0 = vtkClipPolyData()
            clipper_0.SetInputDataObject(cut_surface_polydata)
            clipper_0.SetClipFunction(plane_0)
            clipper_0.Update()
            clipped_polydata = clipper_0.GetOutput()

            clipper_1 = vtkClipPolyData()
            clipper_1.SetInputDataObject(clipped_polydata)
            clipper_1.SetClipFunction(plane_1)
            clipper_1.Update()
            clipped_polydata = clipper_1.GetOutput()

            append_alg.AddInputData(clipped_polydata)

            et_clip_s += timer.lap_s()

        append_alg.Update()
        comb_polydata = append_alg.GetOutput()
        et_combine_s = timer.lap_s()

        points_np = vtk_to_numpy(comb_polydata.GetPoints().GetData()).ravel()
        polys_np = vtk_to_numpy(comb_polydata.GetPolys().GetData())

        surface_polys = SurfacePolys(point_arr=points_np, poly_arr=polys_np)

        property_scalars: Optional[PropertyScalars] = None
        if property_spec:
            raw_cell_vals = _load_property_values(provider, realization, property_spec)
            if raw_cell_vals is not None:
                original_cell_indices_np = vtk_to_numpy(
                    comb_polydata.GetCellData().GetAbstractArray("vtkOriginalCellIds")
                )
                mapped_cell_vals = raw_cell_vals[original_cell_indices_np]
                property_scalars = PropertyScalars(value_arr=mapped_cell_vals)

        LOGGER.debug(
            f"Cutting along polyline done in {timer.elapsed_s():.2f}s "
            f"setup={et_setup_s:.2f}s, cut={et_cut_s:.2f}s, clip={et_clip_s:.2f}s "
            f"combine={et_combine_s:.2f}s, "
            f"(provider_id={provider_id}, real={realization})"
        )

        return surface_polys, property_scalars

    # -----------------------------------------------------------------------------
    # pylint: disable=too-many-locals,
    def ray_pick(
        self,
        provider_id: str,
        realization: int,
        ray: Ray,
        property_spec: Optional[PropertySpec],
        cell_filter: Optional[CellFilter],
    ) -> Optional[PickResult]:

        LOGGER.debug(
            f"Doing ray pick: "
            f"ray.origin={ray.origin}, ray.end={ray.end}, "
            f"(provider_id={provider_id}, real={realization}, "
            f"{_property_spec_dbg_str(property_spec)}, "
            f"{_cell_filter_dbg_str(cell_filter)})"
        )
        timer = PerfTimer()

        provider = self._id_to_provider_dict.get(provider_id)
        if not provider:
            raise ValueError("Could not find provider")

        worker = self._get_or_create_grid_worker(provider_id, realization)
        if not worker:
            raise ValueError("Could not get grid worker")

        grid = worker.get_full_esgrid()
        if cell_filter:
            grid = _calc_cropped_grid(grid, cell_filter)
        et_crop_s = timer.lap_s()

        cell_id, isect_pt = _raypick_in_grid(grid, ray)  # type:ignore
        et_pick_s = timer.lap_s()
        if cell_id is None:
            return None

        original_cell_id = cell_id
        if cell_filter:
            # If a cell filter is present, assume picking was done against cropped grid
            original_cell_id = (
                grid.GetCellData()
                .GetAbstractArray("vtkOriginalCellIds")
                .GetValue(cell_id)
            )

        i_ref = reference(0)
        j_ref = reference(0)
        k_ref = reference(0)
        grid.ComputeCellStructuredCoords(cell_id, i_ref, j_ref, k_ref, True)

        cell_property_val: Optional[float] = None
        if property_spec:
            raw_cell_vals = _load_property_values(provider, realization, property_spec)
            if raw_cell_vals is not None:
                cell_property_val = raw_cell_vals[original_cell_id]
        et_props_s = timer.lap_s()

        LOGGER.debug(
            f"Did ray pick in {timer.elapsed_s():.2f}s ("
            f"crop={et_crop_s:.2f}s, pick={et_pick_s:.2f}s, props={et_props_s:.2f}s, "
            f"provider_id={provider_id}, real={realization}, "
            f"{_property_spec_dbg_str(property_spec)}, "
            f"{_cell_filter_dbg_str(cell_filter)})"
        )

        return PickResult(
            cell_index=original_cell_id,
            cell_i=i_ref.get(),
            cell_j=j_ref.get(),
            cell_k=k_ref.get(),
            intersection_point=isect_pt,
            cell_property_value=cell_property_val,
        )

    # -----------------------------------------------------------------------------
    def _get_or_create_grid_worker(
        self, provider_id: str, realization: int
    ) -> Optional[GridWorker]:

        timer = PerfTimer()

        worker_key = f"P{provider_id}__R{realization}"
        worker = self._key_to_worker_dict.get(worker_key)
        if worker:
            LOGGER.debug("_get_or_create_grid_worker() returning cached data")
            return worker

        provider = self._id_to_provider_dict.get(provider_id)
        if not provider:
            raise ValueError("Could not find provider")

        LOGGER.debug("_get_or_create_grid_worker() data not in cache, loading...")

        xtg_grid = provider.get_3dgrid(realization=realization)
        et_xtgeo_grid_from_provider_grid_ms = timer.lap_ms()

        cell_count = xtg_grid.ncol * xtg_grid.nrow * xtg_grid.nlay
        LOGGER.debug(f"_get_or_create_grid_worker() grid cell count: {cell_count}")

        vtk_esg = xtgeo_grid_to_vtk_explicit_structured_grid(xtg_grid)
        et_create_vtk_esg_ms = timer.lap_ms()

        worker = GridWorker(vtk_esg)
        self._key_to_worker_dict[worker_key] = worker

        LOGGER.debug(
            f"_get_or_create_grid_worker() loaded data in {timer.elapsed_s():.2f}s "
            f"(xtgeo_grid_from_provider_grid={et_xtgeo_grid_from_provider_grid_ms}ms, "
            f"create_vtk_esg={et_create_vtk_esg_ms}ms)"
        )

        return worker


# -----------------------------------------------------------------------------
def _calc_cropped_grid(
    esgrid: vtkExplicitStructuredGrid, cell_filter: CellFilter
) -> vtkExplicitStructuredGrid:
    crop_filter = vtkExplicitStructuredGridCrop()
    crop_filter.SetInputData(esgrid)

    # In VTK dimensions correspond to points
    crop_filter.SetOutputWholeExtent(
        cell_filter.i_min,
        cell_filter.i_max + 1,
        cell_filter.j_min,
        cell_filter.j_max + 1,
        cell_filter.k_min,
        cell_filter.k_max + 1,
    )
    crop_filter.Update()
    cropped_grid = crop_filter.GetOutput()
    return cropped_grid


# -----------------------------------------------------------------------------
def _raypick_in_grid(
    esgrid: vtkExplicitStructuredGrid, ray: Ray
) -> Optional[Tuple[int, List[float]]]:
    """Do a ray pick against the specified grid.
    Returns None if nothing was hit, otherwise returns the cellId (cell index) of the cell
    that was hit and the intersection point
    """

    locator = vtkCellLocator()
    locator.SetDataSet(esgrid)
    locator.BuildLocator()

    tolerance = 0.0
    vtk_isect_points = vtkPoints()
    vtk_cell_ids_list = vtkIdList()

    # Initially we used another vtkGenericCell.IntersectWithLine() overload here that
    # only returned the best hit. However it seems that as of 9.2.0rc2, that function is
    # broken since it seemingly doesn't return the closest hit.
    # For now we try and use another overload that returns all hits, sorted by distance.
    # There is also an overload without the tol argument, but it doesn't seem to work.
    any_hits = locator.IntersectWithLine(
        ray.origin, ray.end, tolerance, vtk_isect_points, vtk_cell_ids_list
    )
    # print(f"{any_hits=}")
    # print(f"{vtk_isect_points.GetPoint(0)}")
    # print(f"{vtk_cell_ids_list.GetId(0)}")

    if any_hits != 1:
        return None

    num_points = vtk_isect_points.GetNumberOfPoints()
    num_ids = vtk_cell_ids_list.GetNumberOfIds()
    if num_points == 0 or num_ids == 0:
        raise ValueError("Raypick got a hit, but no points or Ids were returned")

    isect_pt = list(vtk_isect_points.GetPoint(0))
    cell_id = vtk_cell_ids_list.GetId(0)

    return cell_id, isect_pt


# -----------------------------------------------------------------------------
def _calc_grid_surface(esgrid: vtkExplicitStructuredGrid) -> vtkPolyData:
    surf_filter = vtkExplicitStructuredGridSurfaceFilter()
    surf_filter.SetInputData(esgrid)
    surf_filter.PassThroughCellIdsOn()
    surf_filter.Update()

    polydata: vtkPolyData = surf_filter.GetOutput()
    return polydata


# -----------------------------------------------------------------------------
def _load_property_values(
    provider: EnsembleGridProvider, realization: int, property_spec: PropertySpec
) -> Optional[np.ndarray]:
    timer = PerfTimer()

    if property_spec.prop_date:
        prop_values = provider.get_dynamic_property_values(
            property_spec.prop_name, property_spec.prop_date, realization
        )
    else:
        prop_values = provider.get_static_property_values(
            property_spec.prop_name, realization
        )

    LOGGER.debug(
        f"_load_property_values() took {timer.elapsed_s():.2f}s "
        f"({_property_spec_dbg_str(property_spec)})"
    )

    return prop_values


# -----------------------------------------------------------------------------
def _vtk_esg_to_ug(vtk_esgrid: vtkExplicitStructuredGrid) -> vtkUnstructuredGrid:
    convert_filter = vtkExplicitStructuredGridToUnstructuredGrid()
    convert_filter.SetInputData(vtk_esgrid)
    convert_filter.Update()
    vtk_ugrid = convert_filter.GetOutput()

    return vtk_ugrid


# -----------------------------------------------------------------------------
def _vtk_ug_to_esg(vtk_ugrid: vtkUnstructuredGrid) -> vtkExplicitStructuredGrid:
    convert_filter = vtkUnstructuredGridToExplicitStructuredGrid()
    convert_filter.SetInputData(vtk_ugrid)
    convert_filter.SetInputArrayToProcess(0, 0, 0, 1, "BLOCK_I")
    convert_filter.SetInputArrayToProcess(1, 0, 0, 1, "BLOCK_J")
    convert_filter.SetInputArrayToProcess(2, 0, 0, 1, "BLOCK_K")
    convert_filter.Update()
    vtk_esgrid = convert_filter.GetOutput()

    return vtk_esgrid


# -----------------------------------------------------------------------------
def _calc_polyline_bounds(polyline_xy: List[float]) -> Optional[List[float]]:
    num_points = int(len(polyline_xy) / 2)
    if num_points < 1:
        return None

    min_x = min(polyline_xy[0::2])
    max_x = max(polyline_xy[0::2])
    min_y = min(polyline_xy[1::2])
    max_y = max(polyline_xy[1::2])

    return [min_x, max_x, min_y, max_y]


# -----------------------------------------------------------------------------
# pylint: disable=too-many-locals,
def _extract_intersected_ugrid(
    ugrid: vtkUnstructuredGrid, polyline_xy_in: List[float], max_point_dist: float
) -> vtkUnstructuredGrid:

    timer = PerfTimer()

    polyline_xy = _resample_polyline(polyline_xy_in, max_point_dist)
    et_resample_s = timer.lap_s()

    num_points_in_polyline = int(len(polyline_xy) / 2)
    if num_points_in_polyline < 1:
        return ugrid

    bounds = ugrid.GetBounds()
    min_z = bounds[4]
    max_z = bounds[5]

    points = vtkPoints()
    lines = vtkCellArray()

    for i in range(0, num_points_in_polyline):
        x = polyline_xy[2 * i]
        y = polyline_xy[2 * i + 1]

        points.InsertNextPoint([x, y, min_z])
        points.InsertNextPoint([x, y, max_z])

        line = vtkLine()
        line.GetPointIds().SetId(0, 2 * i)
        line.GetPointIds().SetId(1, 2 * i + 1)

        lines.InsertNextCell(line)

    poly_data = vtkPolyData()
    poly_data.SetPoints(points)
    poly_data.SetLines(lines)

    et_build_s = timer.lap_s()

    extractor = vtkExtractCellsAlongPolyLine()
    extractor.SetInputData(0, ugrid)
    extractor.SetInputData(1, poly_data)
    extractor.Update()

    ret_grid = extractor.GetOutput(0)

    et_extract_s = timer.lap_s()

    LOGGER.debug(
        f"extraction with {num_points_in_polyline} points took {timer.elapsed_s():.2f}s "
        f"(resample={et_resample_s:.2f}s, build={et_build_s:.2f}s, extract={et_extract_s:.2f}s)"
    )

    return ret_grid


# -----------------------------------------------------------------------------
# pylint: disable=too-many-locals,
def _resample_polyline(polyline_xy: List[float], max_point_dist: float) -> List[float]:
    num_points = int(len(polyline_xy) / 2)
    if num_points < 2:
        return polyline_xy

    ret_polyline = []

    prev_x = polyline_xy[0]
    prev_y = polyline_xy[1]
    ret_polyline.extend([prev_x, prev_y])

    for i in range(1, num_points):
        x = polyline_xy[2 * i]
        y = polyline_xy[2 * i + 1]

        fwd = [x - prev_x, y - prev_y]
        length = np.linalg.norm(fwd)
        if length > max_point_dist:
            n_length = int(length / max_point_dist)
            delta_t = 1.0 / (n_length + 1)
            for j in range(0, n_length):
                pt_x = prev_x + fwd[0] * (j + 1) * delta_t
                pt_y = prev_y + fwd[1] * (j + 1) * delta_t
                ret_polyline.extend([pt_x, pt_y])

        ret_polyline.extend([x, y])
        prev_x = x
        prev_y = y

    return ret_polyline


# -----------------------------------------------------------------------------
def _property_spec_dbg_str(property_spec: Optional[PropertySpec]) -> str:
    if not property_spec:
        return "prop=None"

    return f"prop=({property_spec.prop_name}, {property_spec.prop_date})"


# -----------------------------------------------------------------------------
def _cell_filter_dbg_str(cell_filter: Optional[CellFilter]) -> str:
    if not cell_filter:
        return "IJK=None"

    return (
        f"I=[{cell_filter.i_min},{cell_filter.i_max}] "
        f"J=[{cell_filter.j_min},{cell_filter.j_max}] "
        f"K=[{cell_filter.k_min},{cell_filter.k_max}]"
    )



from typing import Any, Dict, Optional

import xtgeo


def xtgeo_well_logs_to_json_format(well: xtgeo.Well) -> Dict:

    header = generate_header(well_name=well.name)
    curves = []

    # Calculate well geometrics if MD log is not provided
    if well.mdlogname is None:
        well.geometrics()

    # Add MD and TVD curves
    curves.append(generate_curve(log_name="MD", description="Measured depth"))
    curves.append(
        generate_curve(log_name="TVD", description="True vertical depth (SS)")
    )
    # Add additonal logs, skipping geometrical logs if calculated
    lognames = [
        logname
        for logname in well.lognames
        if logname not in ["Q_MDEPTH", "Q_AZI", "Q_INCL", "R_HLEN"]
    ]
    for logname in lognames:
        curves.append(generate_curve(log_name=logname.upper()))

    # Filter dataframe to only include relevant logs
    curve_names = [well.mdlogname, "Z_TVDSS"] + lognames
    dframe = well.dataframe[curve_names]
    dframe = dframe.reindex(curve_names, axis=1)

    return {"header": header, "curves": curves, "data": dframe.values.tolist()}


def generate_header(well_name: str, logrun_name: str = "log") -> Dict[str, Any]:
    return {
        "name": logrun_name,
        "well": well_name,
    }


def generate_curve(
    log_name: str, description: Optional[str] = None, value_type: str = "float"
) -> Dict[str, Any]:
    return {
        "name": log_name,
        "description": description,
        "valueType": value_type,
        "dimensions": 1,
        "unit": "m",
        "quantity": None,
        "axis": None,
        "maxSize": 20,
    }



from typing import Dict, List, Tuple, Union

import pandas as pd
import webviz_core_components as wcc
from dash import ALL, Input, Output, callback
from dash.development.base_component import Component
from webviz_config import WebvizSettings
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import ViewABC

from ._settings import DataSettings, ViewSettings
from ._utils._plot_utils import ColorBy, create_graph, filter_data_frame


class PvtView(ViewABC):
    class Ids(StrEnum):
        PVT_GRAPHS = "formation-volume-factor"
        DATA_SETTINGS = "DataSettings"
        SHOW_PLOTS = "show-plots"

    PHASES = ["OIL", "GAS", "WATER"]

    def __init__(self, pvt_df: pd.DataFrame, webviz_settings: WebvizSettings) -> None:
        super().__init__("Pvt View")

        self.pvt_df = pvt_df
        self.plotly_theme = webviz_settings.theme.plotly_theme

        self.phases_additional_info: List[str] = [
            keyword
            for keyword in ["PVTO", "PVDO", "PVCDO", "PVTG", "PVDG", "PVTW"]
            if self.pvt_df["KEYWORD"].str.contains(keyword).any()
        ]

        self.add_settings_groups(
            {
                PvtView.Ids.DATA_SETTINGS: DataSettings(self.pvt_df),
                PvtView.Ids.SHOW_PLOTS: ViewSettings(),
            }
        )

        self.add_column(PvtView.Ids.PVT_GRAPHS)

    @property
    def phases(self) -> Dict[str, str]:
        phase_descriptions: Dict[str, str] = {}
        for i, phase in enumerate(PvtView.PHASES):
            phase_descriptions[phase] = self.phases_additional_info[i]
        return phase_descriptions

    @property
    def ensembles(self) -> List[str]:
        return list(self.pvt_df["ENSEMBLE"].unique())

    @property
    def pvtnums(self) -> List[str]:
        return list(self.pvt_df["PVTNUM"].unique())

    @property
    def ensemble_colors(self) -> Dict[str, List[str]]:
        return {
            ensemble: self.plotly_theme["layout"]["colorway"][
                self.ensembles.index(ensemble)
            ]
            for ensemble in self.ensembles
        }

    @property
    def pvtnum_colors(self) -> Dict[str, List[str]]:
        return {
            pvtnum: self.plotly_theme["layout"]["colorway"][self.pvtnums.index(pvtnum)]
            for pvtnum in self.pvtnums
        }

    def set_callbacks(self) -> None:
        @callback(
            [
                Output(
                    {
                        "id": self.settings_group(PvtView.Ids.SHOW_PLOTS)
                        .component_unique_id(ViewSettings.Ids.SHOW_PLOTS)
                        .to_string(),
                        "plot": plot_value,
                    },
                    "style",
                )
                for plot_value in ViewSettings.plot_visibility_options()
            ],
            Input(
                self.settings_group(PvtView.Ids.DATA_SETTINGS)
                .component_unique_id(DataSettings.Ids.PHASE)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _set_available_plots(
            phase: str,
        ) -> Tuple[dict, ...]:
            all_visibility_options = ViewSettings.plot_visibility_options()
            visibility_options = ViewSettings.plot_visibility_options(phase)
            return tuple(
                {"display": "block" if plot in visibility_options else "none"}
                for plot in all_visibility_options
            )

        @callback(
            Output(
                self.layout_element(PvtView.Ids.PVT_GRAPHS).get_unique_id().to_string(),
                "children",
            ),
            Input(
                self.settings_group(PvtView.Ids.DATA_SETTINGS)
                .component_unique_id(DataSettings.Ids.COLOR_BY)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(PvtView.Ids.DATA_SETTINGS)
                .component_unique_id(DataSettings.Ids.ENSEMBLES)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(PvtView.Ids.DATA_SETTINGS)
                .component_unique_id(DataSettings.Ids.PHASE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(PvtView.Ids.DATA_SETTINGS)
                .component_unique_id(DataSettings.Ids.PVTNUM)
                .to_string(),
                "value",
            ),
            Input(
                {
                    "id": self.settings_group(PvtView.Ids.SHOW_PLOTS)
                    .component_unique_id(ViewSettings.Ids.SHOW_PLOTS)
                    .to_string(),
                    "plot": ALL,
                },
                "id",
            ),
            Input(
                {
                    "id": self.settings_group(PvtView.Ids.SHOW_PLOTS)
                    .component_unique_id(ViewSettings.Ids.SHOW_PLOTS)
                    .to_string(),
                    "plot": ALL,
                },
                "value",
            ),
        )
        @callback_typecheck
        # pylint: disable=too-many-locals
        def _update_plots(
            color_by: ColorBy,
            selected_ensembles: Union[List[str], str],
            phase: str,
            selected_pvtnum: Union[List[int], int],
            plots_visibility_ids: List[dict],
            plots_visibility: List[str],
        ) -> List[Component]:

            if isinstance(selected_ensembles, str):
                ensembles = [selected_ensembles]
            else:
                ensembles = selected_ensembles
            if isinstance(selected_pvtnum, int):
                pvtnum = [selected_pvtnum]
            else:
                pvtnum = selected_pvtnum

            pvt_df = filter_data_frame(self.pvt_df, ensembles, pvtnum)

            if color_by == ColorBy.ENSEMBLE:
                colors = self.ensemble_colors
            elif color_by == ColorBy.PVTNUM:
                colors = self.pvtnum_colors

            max_num_columns = 2
            view_elements = []

            current_row_elements = []
            current_column = 0

            current_column = 0
            figure_index = 0

            visible_plots = {
                plot_id["plot"]: plot_id["plot"] in visibility
                for plot_id, visibility in zip(plots_visibility_ids, plots_visibility)
            }

            graph_height = max(
                40.0,
                80.0
                / max(
                    len(
                        [
                            plot
                            for plot in ViewSettings.plot_visibility_options(phase)
                            if visible_plots[plot]
                        ]
                    ),
                    1,
                ),
            )

            for plot in ViewSettings.plot_visibility_options(phase):
                if not visible_plots[plot]:
                    continue

                current_element = wcc.WebvizViewElement(
                    id=self.unique_id(plot),
                    children=create_graph(
                        pvt_df,
                        color_by,
                        colors,
                        phase,
                        plot,
                        ViewSettings.plot_visibility_options(self.phases[phase])[plot],
                        self.plotly_theme,
                        graph_height,
                    ),
                )
                current_row_elements.append(current_element)

                current_column += 1

                if current_column >= max_num_columns:
                    view_elements.append(
                        wcc.WebvizPluginLayoutRow(current_row_elements)
                    )
                    current_row_elements = []
                    current_column = 0

                figure_index += 1

            if len(current_row_elements) > 0:
                view_elements.append(wcc.WebvizPluginLayoutRow(current_row_elements))

            return view_elements



import datetime

import pandas as pd
from pandas._testing import assert_frame_equal

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.dataframe_utils import (
    create_relative_to_date_df,
)


def test_create_relative_to_date_df_consistent_realization_for_all_dates() -> None:
    # All dates are represented across all realizations
    input_df = pd.DataFrame(
        columns=["DATE", "REAL", "A", "B"],
        data=[
            [datetime.datetime(2263, 1, 1), 1, 10.0, 130.0],
            [datetime.datetime(2263, 2, 1), 1, 45.0, 135.0],
            [datetime.datetime(2263, 3, 1), 1, 50.0, 140.0],
            [datetime.datetime(2263, 4, 1), 1, 55.0, 145.0],
            [datetime.datetime(2263, 1, 1), 2, 11.0, 150.0],
            [datetime.datetime(2263, 2, 1), 2, 65.0, 155.0],
            [datetime.datetime(2263, 3, 1), 2, 70.0, 160.0],
            [datetime.datetime(2263, 4, 1), 2, 75.0, 165.0],
            [datetime.datetime(2263, 1, 1), 3, 12.0, 170.0],
            [datetime.datetime(2263, 2, 1), 3, 85.0, 175.0],
            [datetime.datetime(2263, 3, 1), 3, 90.0, 180.0],
            [datetime.datetime(2263, 4, 1), 3, 95.0, 185.0],
        ],
    )
    expected_df = pd.DataFrame(
        columns=["DATE", "REAL", "A", "B"],
        data=[
            [datetime.datetime(2263, 1, 1), 1, 0.0, 0.0],
            [datetime.datetime(2263, 2, 1), 1, 35.0, 5.0],
            [datetime.datetime(2263, 3, 1), 1, 40.0, 10.0],
            [datetime.datetime(2263, 4, 1), 1, 45.0, 15.0],
            [datetime.datetime(2263, 1, 1), 2, 0.0, 0.0],
            [datetime.datetime(2263, 2, 1), 2, 54.0, 5.0],
            [datetime.datetime(2263, 3, 1), 2, 59.0, 10.0],
            [datetime.datetime(2263, 4, 1), 2, 64.0, 15.0],
            [datetime.datetime(2263, 1, 1), 3, 0.0, 0.0],
            [datetime.datetime(2263, 2, 1), 3, 73.0, 5.0],
            [datetime.datetime(2263, 3, 1), 3, 78.0, 10.0],
            [datetime.datetime(2263, 4, 1), 3, 83.0, 15.0],
        ],
    )

    assert_frame_equal(
        create_relative_to_date_df(input_df, datetime.datetime(2263, 1, 1)), expected_df
    )


def test_create_relative_to_date_df_relative_date_missing_a_realization() -> None:
    # Missing datetime.datetime(2263, 1, 1) for real = 3!
    input_df = pd.DataFrame(
        columns=["DATE", "REAL", "A", "B"],
        data=[
            [datetime.datetime(2263, 1, 1), 1, 10.0, 130.0],
            [datetime.datetime(2263, 2, 1), 1, 45.0, 135.0],
            [datetime.datetime(2263, 3, 1), 1, 50.0, 140.0],
            [datetime.datetime(2263, 4, 1), 1, 55.0, 145.0],
            [datetime.datetime(2263, 1, 1), 2, 11.0, 150.0],
            [datetime.datetime(2263, 2, 1), 2, 65.0, 155.0],
            [datetime.datetime(2263, 3, 1), 2, 70.0, 160.0],
            [datetime.datetime(2263, 4, 1), 2, 75.0, 165.0],
            [datetime.datetime(2263, 2, 1), 3, 85.0, 175.0],
            [datetime.datetime(2263, 3, 1), 3, 90.0, 180.0],
            [datetime.datetime(2263, 4, 1), 3, 95.0, 185.0],
        ],
    )
    expected_df = pd.DataFrame(
        columns=["DATE", "REAL", "A", "B"],
        data=[
            [datetime.datetime(2263, 1, 1), 1, 0.0, 0.0],
            [datetime.datetime(2263, 2, 1), 1, 35.0, 5.0],
            [datetime.datetime(2263, 3, 1), 1, 40.0, 10.0],
            [datetime.datetime(2263, 4, 1), 1, 45.0, 15.0],
            [datetime.datetime(2263, 1, 1), 2, 0.0, 0.0],
            [datetime.datetime(2263, 2, 1), 2, 54.0, 5.0],
            [datetime.datetime(2263, 3, 1), 2, 59.0, 10.0],
            [datetime.datetime(2263, 4, 1), 2, 64.0, 15.0],
        ],
    )

    assert_frame_equal(
        create_relative_to_date_df(input_df, datetime.datetime(2263, 1, 1)), expected_df
    )


def test_create_relative_to_date_df_relative_date_not_existing() -> None:
    # Missing datetime.datetime(2263, 5, 1)
    input_df = pd.DataFrame(
        columns=["DATE", "REAL", "A", "B"],
        data=[
            [datetime.datetime(2263, 1, 1), 1, 10.0, 130.0],
            [datetime.datetime(2263, 2, 1), 1, 45.0, 135.0],
            [datetime.datetime(2263, 3, 1), 1, 50.0, 140.0],
            [datetime.datetime(2263, 4, 1), 1, 55.0, 145.0],
            [datetime.datetime(2263, 1, 1), 2, 11.0, 150.0],
            [datetime.datetime(2263, 2, 1), 2, 65.0, 155.0],
            [datetime.datetime(2263, 3, 1), 2, 70.0, 160.0],
            [datetime.datetime(2263, 4, 1), 2, 75.0, 165.0],
            [datetime.datetime(2263, 1, 1), 3, 12.0, 170.0],
            [datetime.datetime(2263, 2, 1), 3, 85.0, 175.0],
            [datetime.datetime(2263, 3, 1), 3, 90.0, 180.0],
            [datetime.datetime(2263, 4, 1), 3, 95.0, 185.0],
        ],
    )

    # Ensure dtype for columns for df with no rows
    _columns = {
        name: pd.Series(dtype=input_df.dtypes[name]) for name in input_df.columns
    }
    expected_df = pd.DataFrame(_columns)

    assert_frame_equal(
        create_relative_to_date_df(input_df, datetime.datetime(2263, 5, 1)), expected_df
    )



from enum import Enum
from typing import Callable

import webviz_core_components as wcc
from dash import html
from webviz_config import WebvizConfigTheme

from webviz_subsurface._components.parameter_filter import ParameterFilter

from ..models import ParametersModel, SimulationTimeSeriesModel
from .selector_view import (
    button,
    color_opacity_selector,
    color_selector,
    date_selector,
    ensemble_selector,
    filter_vector_selector,
    parameter_selector,
    plot_options,
    vector_selector,
)


class VisualizationOptions(str, Enum):
    """
    Type definition for visualization options in simulation time series
    """

    REALIZATIONS = "realizations"
    STATISTICS = "statistics"
    STATISTICS_AND_REALIZATIONS = "statistics and realizations"


def timeseries_view(get_uuid: Callable) -> html.Div:
    return html.Div(
        children=[
            wcc.Graph(
                id=get_uuid("vector-vs-time-graph"),
                config={"displayModeBar": False},
                style={"height": "38vh"},
            ),
        ],
    )


def selector_view(
    get_uuid: Callable,
    vectormodel: SimulationTimeSeriesModel,
    parametermodel: ParametersModel,
    theme: WebvizConfigTheme,
) -> wcc.Frame:

    theme_colors = theme.plotly_theme.get("layout", {}).get("colorway", [])
    return wcc.Frame(
        style={
            "height": "80vh",
            "overflowY": "auto",
            "font-size": "15px",
        },
        children=[
            wcc.Selectors(
                label="Selections",
                children=[
                    ensemble_selector(
                        get_uuid=get_uuid,
                        ensembles=parametermodel.mc_ensembles,
                        tab="response",
                        id_string="ensemble-selector",
                        heading="Ensemble:",
                    ),
                    vector_selector(get_uuid=get_uuid, vectormodel=vectormodel),
                    date_selector(get_uuid=get_uuid, vectormodel=vectormodel),
                    parameter_selector(
                        get_uuid=get_uuid, parametermodel=parametermodel, tab="response"
                    ),
                ],
            ),
            wcc.Selectors(
                label="Visualization",
                children=[
                    wcc.RadioItems(
                        id=get_uuid("visualization"),
                        options=[
                            {
                                "label": "Individual realizations",
                                "value": VisualizationOptions.REALIZATIONS.value,
                            },
                            {
                                "label": "Statistical lines",
                                "value": VisualizationOptions.STATISTICS.value,
                            },
                            {
                                "label": "Statistics + Realizations",
                                "value": VisualizationOptions.STATISTICS_AND_REALIZATIONS,
                            },
                        ],
                        value=VisualizationOptions.REALIZATIONS.value,
                    )
                ],
            ),
            wcc.Selectors(
                label="Vectors for parameter correlation",
                children=filter_vector_selector(get_uuid=get_uuid),
            ),
            wcc.Selectors(
                label="Options",
                children=[
                    parameter_filter_button(get_uuid),
                    options_layout(get_uuid, theme_colors),
                ],
            ),
        ],
    )


def parameter_filter_button(get_uuid: Callable):
    return button(get_uuid("display-paramfilter"), "Show parameter filter")


def options_layout(get_uuid: Callable, theme_colors: list):
    return html.Div(
        children=[
            button(get_uuid("options-button"), "Options"),
            wcc.Dialog(
                title="Options",
                id=get_uuid("options-dialog"),
                backdrop=False,
                draggable=True,
                open=False,
                children=[
                    plot_options(get_uuid=get_uuid, tab="response"),
                    color_selector(
                        get_uuid=get_uuid,
                        tab="response",
                        colors=[theme_colors, "Greys", "BrBG"],
                        bargap=0.2,
                        height=50,
                    ),
                    color_opacity_selector(
                        get_uuid=get_uuid, tab="response", value=0.5
                    ),
                ],
            ),
        ]
    )


def parameter_response_view(
    get_uuid: Callable,
    parametermodel: ParametersModel,
    vectormodel: SimulationTimeSeriesModel,
    theme: WebvizConfigTheme,
) -> wcc.FlexBox:
    df = parametermodel.dataframe
    parameter_filter = ParameterFilter(
        uuid=get_uuid("parameter-filter"),
        dframe=df[df["ENSEMBLE"].isin(parametermodel.mc_ensembles)].copy(),
        reset_on_ensemble_update=True,
    )
    return wcc.FlexBox(
        children=[
            wcc.FlexColumn(
                flex=1,
                children=selector_view(
                    get_uuid=get_uuid,
                    parametermodel=parametermodel,
                    vectormodel=vectormodel,
                    theme=theme,
                ),
            ),
            wcc.FlexColumn(
                flex=4,
                children=wcc.FlexBox(
                    children=[
                        wcc.FlexColumn(
                            flex=2,
                            children=[
                                wcc.Frame(
                                    style={"height": "38.5vh"},
                                    color="white",
                                    highlight=False,
                                    children=timeseries_view(get_uuid=get_uuid),
                                ),
                                wcc.Frame(
                                    style={"height": "38.5vh"},
                                    color="white",
                                    highlight=False,
                                    children=[
                                        wcc.Graph(
                                            id=get_uuid("vector-vs-param-scatter"),
                                            config={"displayModeBar": False},
                                            style={"height": "38vh"},
                                        )
                                    ],
                                ),
                            ],
                        ),
                        wcc.FlexColumn(
                            flex=2,
                            children=[
                                wcc.Frame(
                                    color="white",
                                    highlight=False,
                                    style={"height": "38.5vh"},
                                    children=[
                                        wcc.Graph(
                                            config={"displayModeBar": False},
                                            style={"height": "38vh"},
                                            id=get_uuid("vector-corr-graph"),
                                        ),
                                    ],
                                ),
                                wcc.Frame(
                                    color="white",
                                    highlight=False,
                                    style={"height": "38.5vh"},
                                    children=[
                                        wcc.Graph(
                                            config={"displayModeBar": False},
                                            style={"height": "38vh"},
                                            id=get_uuid("param-corr-graph"),
                                        ),
                                    ],
                                ),
                            ],
                        ),
                        wcc.FlexColumn(
                            id=get_uuid("param-filter-wrapper"),
                            style={"display": "none"},
                            flex=1,
                            children=wcc.Frame(
                                style={"height": "80vh"},
                                children=parameter_filter.layout,
                            ),
                        ),
                    ],
                ),
            ),
        ],
    )



import fnmatch
import re
from typing import List, Optional

import pandas as pd

from webviz_subsurface._abbreviations.reservoir_simulation import historical_vector
from webviz_subsurface._utils.dataframe_utils import make_date_column_datetime_object
from webviz_subsurface._utils.simulation_timeseries import (
    set_simulation_line_shape_fallback,
)
from webviz_subsurface._utils.vector_selector import add_vector_to_vector_selector_data


class SimulationTimeSeriesModel:
    """Class to process and and visualize ensemble timeseries"""

    REQUIRED_COLUMNS = ["ENSEMBLE", "REAL", "DATE"]

    def __init__(
        self,
        dataframe: pd.DataFrame,
        line_shape_fallback: str = "linear",
    ) -> None:

        for column in self.REQUIRED_COLUMNS:
            if column not in dataframe.columns:
                raise KeyError(f"{column} column is missing from UNSMRY data")

        dataframe = dataframe.copy()
        # ensure correct format of date
        dataframe["REAL"] = dataframe["REAL"].astype(int)
        dataframe["DATE"] = pd.to_datetime(dataframe["DATE"])
        make_date_column_datetime_object(dataframe)

        self._dataframe = self._remove_columns_with_only_0(dataframe)
        self.line_shape_fallback = set_simulation_line_shape_fallback(
            line_shape_fallback
        )
        self._dates = sorted(self._dataframe["DATE"].unique())
        self._vector_names = self._determine_vector_names()

        # add vectors to vector selector
        self.vector_selector_data: list = []
        for vector in self._vector_names:
            add_vector_to_vector_selector_data(self.vector_selector_data, vector)

    def _determine_vector_names(self) -> List[str]:
        """Determine which vectors we should make available"""
        return [c for c in self._dataframe if c not in self.REQUIRED_COLUMNS]

    @staticmethod
    def _remove_columns_with_only_0(dframe: pd.DataFrame) -> pd.DataFrame:
        """Filter list of vector names by removing vector names where the data is all zeros"""
        return dframe.loc[:, (dframe != 0).any(axis=0)]

    @property
    def dates(self) -> List[str]:
        return self._dates

    @property
    def dataframe(self) -> List[str]:
        return self._dataframe

    @property
    def vectors(self) -> List[str]:
        return self._vector_names

    @property
    def ensembles(self) -> List[str]:
        return list(self._dataframe["ENSEMBLE"].unique())

    def filter_vectors(self, column_keys: str):
        """Filter vector list used for correlation"""
        column_keys = "".join(column_keys.split()).split(",")
        try:
            regex = re.compile(
                "|".join([fnmatch.translate(col) for col in column_keys]),
                flags=re.IGNORECASE,
            )
            return [v for v in self.vectors if regex.fullmatch(v)]
        except re.error:
            return []

    def get_vector_df(
        self,
        ensemble: str,
        realizations: list,
        vectors: Optional[list] = None,
    ) -> pd.DataFrame:
        vectors = vectors if vectors is not None else self._vector_names
        df = self.dataframe[self.REQUIRED_COLUMNS + vectors].copy()
        return df.loc[
            (df["ENSEMBLE"] == ensemble) & (df["REAL"].isin(realizations)),
            ["DATE", "REAL"] + vectors,
        ]

    def get_historical_vector_df(
        self, vector: str, ensemble: str
    ) -> Optional[pd.DataFrame]:
        df = self._dataframe
        hist_vecname = historical_vector(vector, None)
        if hist_vecname and hist_vecname in df.columns:
            return (
                df[[hist_vecname, "DATE"]]
                .loc[
                    (df["REAL"] == df["REAL"].unique()[0])
                    & (df["ENSEMBLE"] == ensemble)
                ]
                .rename(columns={hist_vecname: vector})
            )
        return None



from typing import Dict, List, Optional

import pandas as pd

from webviz_subsurface._abbreviations.reservoir_simulation import historical_vector
from webviz_subsurface._providers import EnsembleSummaryProvider, Frequency


def create_history_vectors_df(
    provider: EnsembleSummaryProvider,
    vector_names: List[str],
    resampling_frequency: Optional[Frequency],
) -> pd.DataFrame:
    """Get dataframe with existing historical vector data for provided vectors.

    The returned dataframe contains columns with name of vector and corresponding historical
    data

    `Input:`
    * ensemble: str - Ensemble name
    * vector_names: List[str] - list of vectors to get historical data for
    [vector1, ... , vectorN]

    `Output:`
    * dataframe with non-historical vector names in columns and their historical data in rows.
    `Columns` in dataframe: ["DATE", "REAL", vector1, ..., vectorN]

    ---------------------
    `NOTE:`
    * Raise ValueError if vector does not exist for ensemble
    * If historical data does not exist for provided vector, vector is excluded from
    the returned dataframe.
    * Column names are not the historical vector name, but the original vector name,
    i.e. `WOPTH:OP_1` data is placed in colum with name `WOPT:OP_1`
    """
    if len(vector_names) < 1:
        raise ValueError("Empty list of vector names!")

    provider_vectors = provider.vector_names()
    resampling_frequency = (
        resampling_frequency if provider.supports_resampling() else None
    )

    # Verify for provider
    for elm in vector_names:
        if elm not in provider_vectors:
            raise ValueError(f'Vector "{elm}" not present among vectors for provider')

    # Dict with historical vector name as key, and non-historical vector name as value
    historical_vector_and_vector_name_dict: Dict[str, str] = {}
    for vector in vector_names:
        # TODO: Create new historical_vector according to new provider metadata?
        historical_vector_name = historical_vector(vector=vector, smry_meta=None)
        if historical_vector_name and historical_vector_name in provider.vector_names():
            historical_vector_and_vector_name_dict[historical_vector_name] = vector

    # Get lowest valid realization number
    realization = min(provider.realizations(), default=None)
    if not historical_vector_and_vector_name_dict or realization is None:
        return pd.DataFrame()

    historical_vector_names = list(historical_vector_and_vector_name_dict.keys())
    historical_vectors_df = provider.get_vectors_df(
        historical_vector_names, resampling_frequency, realizations=[realization]
    )
    return historical_vectors_df.rename(columns=historical_vector_and_vector_name_dict)
